/**************************************
 * $Author: kaneda $
 * $Date: 2003/01/18 10:50:14 $
 **************************************/

/* msg.c */
#include "ph_msg.h"


ph_msg_t ph_msg_new(ph_msg_type_t type, ph_vp_t dest, void * body, size_t len, int tag) {
  ph_msg_t msg;

  msg = (ph_msg_t)ph_malloc(sizeof(struct ph_msg));

  msg->type = type;
  msg->dest = dest;
  msg->body = body;
  msg->len = len;
  msg->tag = tag;

  msg->to_conn = NULL;
  msg->from_conn = NULL;
  msg->meta_type = invalid;
  ph_pthread_mutex_init(&msg->mp, NULL);

  return msg;
}


void ph_msg_dispose(ph_msg_t msg) { 
  if (msg == NULL) {
    ph_err_msg("ph_msg_dispose() : NULL input");
    return;
  }

  ph_pthread_mutex_destroy(&msg->mp);  

  if (msg->body != NULL) { ph_free(msg->body); }
  ph_free(msg);
}


static void ph_msg_print_sub(ph_msg_t msg, void * arg) {
  ph_debug_print("(%p) type=%s, dest=" PH_LLD ", len=%d, tag=%d, conn=%p\n",
		 msg,
		 ((msg->type == meta) ? "meta" : "user"),
		 msg->dest,
		 msg->len,
		 msg->tag,
		 msg->to_conn);
}

void ph_msg_print(ph_msg_t msg) {
  assert(msg != NULL);
  ph_debug_print2("[msg \n");
  ph_msg_print_sub(msg, NULL);
  ph_debug_print2("]\n");
}


ph_msg_t ph_msg_new_user(ph_vp_t dest, void * body, size_t len, int tag) {
  return ph_msg_new(user, dest, body, len, tag);
}


ph_msg_t ph_msg_new_meta(ph_conn_t conn, ph_msg_meta_type_t meta_type, void * body, size_t len) {
  ph_msg_t msg;

  msg = ph_msg_new(meta, ph_get_ph_vp_t(0), body, len, 0);
  msg->meta_type = meta_type;
  msg->to_conn = conn;
  return msg;
}


static void create_header(ph_msg_t msg, char header[]) {
  ph_mzero(header, PH_MSG_HEADER_SIZE);
  ph_snprintf(header, PH_MSG_HEADER_SIZE, 
	      "PHOENIX    TYPE: %s   DEST: " PH_LLD "   MTYPE: %d   TAG: %d   SIZE: %d ",
	      ((msg->type == meta) ? "meta" : "user"),
	      msg->dest,
	      msg->meta_type,
	      msg->tag,
	      msg->len);  
}


static ph_msg_t analyze_header(char header[]) {
  int i;
  ph_msg_t msg;
  char s[6][128];
  char sign[128], type[128];
  ph_vp_t dest;
  int meta_type;
  size_t len;
  int tag;

  ph_mzero(sign, 128);
  ph_mzero(type, 128);
  for (i = 0; i < 6; i++) {
    ph_mzero(s[i], 128);
  }

  sscanf(header, "%s %s%s %s" PH_LLD " %s%d %s%d %s%d",
	 sign,
	 s[0], type, 
	 s[1], &dest,
	 s[2], &meta_type,
	 s[3], &tag,
	 s[4], &len);
  
  if (ph_streq(sign, "PHOENIX") == 0) {
    return NULL;
  }

  msg = ph_msg_new((ph_streq(type, "meta")) ? meta : user,
		   dest, NULL, len, tag);
  msg->meta_type = meta_type;
  
  return msg;
}

static int is_send_success(ssize_t len, ssize_t n) {
  return (n == len);
}


/* if message is sent correctly, then return 1
   otherwise, return 0
 */ 
int ph_msg_send(ph_msg_t msg) {
  char header[PH_MSG_HEADER_SIZE];
  ssize_t n;

  assert(msg != NULL);
  assert(msg->to_conn != NULL);
  assert(msg->body != NULL);
  assert(msg->len > 0);

  create_header(msg, header);

  n = ph_conn_send(msg->to_conn, header, PH_MSG_HEADER_SIZE);

  if (is_send_success(PH_MSG_HEADER_SIZE, n) == 0) {
    return 0;
  }

  n = ph_conn_send(msg->to_conn, msg->body, msg->len);

  if (is_send_success(msg->len, n) == 0) {
    return 0;
  }

  return 1;
}


ph_msg_t ph_msg_recv(ph_conn_t conn) {
  char header[PH_MSG_HEADER_SIZE];
  ssize_t n;
  ph_msg_t msg;
  
  ph_mzero(header, PH_MSG_HEADER_SIZE);
  if ((n = ph_conn_recv(conn, header, PH_MSG_HEADER_SIZE)) <= 0) {
    return NULL;
  }

  if ((msg = analyze_header(header)) == NULL) {
    return NULL;
  }

  msg->from_conn = conn;
  msg->body = (char *)ph_calloc(msg->len, sizeof(char));

  if ((n = ph_conn_recv(conn, msg->body, msg->len)) <= 0) {
    ph_msg_dispose(msg);
    return NULL;
  }
  return msg;
}


static void ph_msg_store(ph_msg_t msg, int fd) {
  char header[PH_MSG_HEADER_SIZE];

  assert(msg != NULL);
  assert(msg->body != NULL);
  assert(msg->len > 0);
  assert(fd > 0);

  create_header(msg, header);

  write(fd, header, PH_MSG_HEADER_SIZE);
  write(fd, msg->body, msg->len);
}


static ph_msg_t ph_msg_load(int fd) {
  ph_msg_t msg;
  char header[PH_MSG_HEADER_SIZE];
  ssize_t n;

  assert(fd > 0);

  if ((n = read(fd, header, PH_MSG_HEADER_SIZE)) <= 0) {
    return NULL;
  }


  msg = analyze_header(header);
  msg->body = (char *)ph_calloc(msg->len, sizeof(char));

  if ((n = read(fd, msg->body, msg->len)) <= 0) {
    ph_msg_dispose(msg->body);
    return NULL;
  }

  return msg;
}


int ph_msg_is_remote_msg(ph_msg_t msg, fd_set * w_fdset_p) {
  int n;
  ph_conn_t conn;

  ph_pthread_mutex_lock(&msg->mp);
  conn = msg->to_conn;
  n = ((conn != NULL) && (*conn->fd_isset_func_p)(conn->attr, w_fdset_p, PH_FD_TYPE_WRITE));
  ph_pthread_mutex_unlock(&msg->mp);
  return n;
}


static int is_remote_msg(ph_msg_t msg, void * arg1, void * arg2) {
  return ph_msg_is_remote_msg(msg, (fd_set *)arg1);
}


static int ph_msg_is_same_tag(int tag1, int tag2) {
  if (tag2 == 0) {
    return 1;
  } else {
    return (tag1 == tag2);
  }
}

int ph_msg_is_local_user_msg(ph_msg_t msg, ph_vp_set_sync_t vp_set, int tag) { 
  int n;
  ph_pthread_mutex_lock(&msg->mp);
  n = ((msg->type == user) && (ph_vp_set_sync_is_member(vp_set, msg->dest)) && (ph_msg_is_same_tag(msg->tag, tag)));
  ph_pthread_mutex_unlock(&msg->mp);
  return n;
}


static int is_local_user_msg(ph_msg_t msg, void * arg1, void * arg2) {
  return ph_msg_is_local_user_msg(msg, (ph_vp_set_sync_t)arg1, (int)arg2);
}


int ph_msg_is_local_meta_msg(ph_msg_t msg, ph_vp_set_sync_t vp_set) {
  int n;
  ph_pthread_mutex_lock(&msg->mp);
  n = ((msg->type == meta) && (msg->to_conn == NULL));
  ph_pthread_mutex_unlock(&msg->mp);
  return n;
}


static int is_local_meta_msg(ph_msg_t msg, void * arg1, void * arg2) {
  return ph_msg_is_local_meta_msg(msg, (ph_vp_set_sync_t)arg1);
}


PH_DEFINE_LIST_FUNC_BODY(ph_msg) 


void ph_msg_list_print(ph_msg_list_t msg_list) {
  assert(msg_list != NULL);

  if (ph_msg_list_is_empty(msg_list)) {
    ph_debug_print2("[msg_list]\n");
  } else {
    ph_debug_print2("[msg_list \n");
    ph_msg_list_map(msg_list, &ph_msg_print_sub, NULL);
    ph_debug_print2("]\n");
  }
}


ph_msg_t ph_msg_list_dequeue_remote_msg(ph_msg_list_t msg_list, fd_set * w_fdset_p) {
  return ph_msg_list_dequeue_opt(msg_list, 0, &is_remote_msg, w_fdset_p, NULL);
}


ph_msg_t ph_msg_list_dequeue_local_user_msg(ph_msg_list_t msg_list, ph_vp_set_sync_t vp_set, int tag, int is_block) {
  return ph_msg_list_dequeue_opt(msg_list, is_block, &is_local_user_msg, (void *)vp_set, (void *)tag);
}


ph_msg_t ph_msg_list_dequeue_local_meta_msg(ph_msg_list_t msg_list, ph_vp_set_sync_t vp_set) {
  return ph_msg_list_dequeue_opt(msg_list, 0, &is_local_meta_msg, vp_set, NULL);
}


void ph_msg_list_store(ph_msg_list_t msg_list, char * filename) {
  int fd;
  ph_msg_t msg;

  assert(filename != NULL);

  if ((fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC | O_NOCTTY, 0666)) < 0) {
    return;
  }

  while ((msg = ph_msg_list_dequeue_opt(msg_list, 0, NULL, NULL, NULL)) != NULL) {
    switch (msg->type) {
    case meta:
      break;
    case user:
      ph_msg_store(msg, fd);
      break;
    default:
      ph_err_quit("ph_msg_list_store() : no matching");
      break;
    }
    ph_msg_dispose(msg);
  }

  ph_close(fd);
}


void ph_msg_list_load(ph_msg_list_t msg_list, char * filename) {
  int fd;
  ph_msg_t msg;

  assert(filename != NULL);

  if ((fd = open(filename, O_RDONLY | O_NOCTTY)) < 0) {
    return;
  }

  while ((msg = ph_msg_load(fd)) != NULL) {
    ph_msg_list_enqueue(msg_list, msg);
  }
  ph_close(fd);
}


int ph_msg_list_is_empty_user_msg(ph_msg_list_t msg_list) {
  ph_msg_cell_t p;
  int n = 1;

  ph_pthread_mutex_lock(&msg_list->mp);
  for (p = msg_list->head; p != NULL; p = p->next) {
    switch (p->self->type) {
    case meta:
      break;
    case user:
      n = 0;
      break;
    default:
      ph_err_quit("ph_msg_list_is_empty_user_msg() : no matching");
      break;
    }
  }
  ph_pthread_mutex_unlock(&msg_list->mp);

  return n;
}


/* This function re-routes/destroies messages that are to be forwared through a broken link */
void ph_msg_list_clear_conn(ph_msg_list_t msg_list, ph_conn_t conn) {
  ph_msg_cell_t p, next;

  ph_pthread_mutex_lock(&msg_list->mp);
  for (p = msg_list->head; p != NULL; p = next) {
    next = p->next;
    
    ph_pthread_mutex_lock(&p->self->mp);
    if ((p->self->to_conn == NULL) || (p->self->to_conn != conn)) {
      ph_pthread_mutex_unlock(&p->self->mp);
      continue;
    }
    
    switch (p->self->type) {
    case meta:
      /* delete a meta message */
      ph_pthread_mutex_unlock(&p->self->mp);
      ph_msg_list_delete_opt(msg_list, p->self, 0);
      break;
    case user:
      /* re-route a user message */
      p->self->to_conn = NULL;
      ph_pthread_mutex_unlock(&p->self->mp);
      break;
    default:
      ph_err_quit("ph_msg_list_clear_conn() : no matching");
      break;
    }
  }
  ph_pthread_mutex_unlock(&msg_list->mp);
}

