/**************************************
 * $Author: kaneda $
 * $Date: 2003/02/12 12:45:20 $
 **************************************/

/* pt2pt.c */

#include "ph_pt2pt.h"

ph_pt2pt_t ph_pt2pt_new(int listen_port_lower, int listen_port_upper) {
  ph_pt2pt_t pt2pt;
  
  pt2pt = (ph_pt2pt_t)ph_malloc(sizeof(struct ph_pt2pt));

  pt2pt->vp_set = ph_vp_set_sync_new();
  pt2pt->resource_vp = ph_pt2pt_get_temp_vp();
  pt2pt->addr_list = ph_addr_list_new(1);
  pt2pt->routing = ph_routing_new(0);
  pt2pt->conn_list = ph_conn_list_new(1);
  pt2pt->msg_list = ph_msg_list_new(1);
  pt2pt->available_ports[0] = listen_port_lower;
  pt2pt->available_ports[1] = listen_port_upper;
  pt2pt->listen_port = -1;
  pt2pt->is_finalizing = 0;
  pt2pt->is_disposed = 0;
  ph_pthread_mutex_init(&pt2pt->mp, NULL);

  pt2pt->checker = 0;

#ifdef PH_VIEWER
  pt2pt->is_viewing = 0;
  pt2pt->viewer_update_interval = -1;
  pt2pt->viewer_addr = NULL;
  pt2pt->viewer_conn = NULL;
#endif

  return pt2pt;
}


void ph_pt2pt_dispose(ph_pt2pt_t pt2pt) {
  if (pt2pt == NULL) {
    ph_err_msg("ph_pt2pt_dispoes() : NULL input");
    return;
  }
  assert(pt2pt->is_disposed == 1);

  ph_vp_set_sync_dispose(pt2pt->vp_set);
  ph_addr_list_dispose(pt2pt->addr_list);
  ph_routing_dispose(pt2pt->routing);
  ph_conn_list_dispose(pt2pt->conn_list);
  ph_msg_list_dispose(pt2pt->msg_list);
  ph_pthread_mutex_destroy(&pt2pt->mp);

#ifdef PH_VIEWER
  if (pt2pt->viewer_addr != NULL) { ph_addr_dispose(pt2pt->viewer_addr); }
  if (pt2pt->viewer_conn != NULL) { ph_conn_dispose(pt2pt->viewer_conn); }
#endif

  ph_free(pt2pt);
}


void ph_pt2pt_print(ph_pt2pt_t pt2pt) {
  assert(pt2pt != NULL);

  if (ph_debug_get_mode() == 0) {
    return;
  }

  ph_debug_print2("\n[pt2pt ");
  ph_debug_print("listen_port=%d\n", pt2pt->listen_port);
  ph_vp_set_print(pt2pt->vp_set->vp_set);
  //ph_addr_list_print(pt2pt->addr_list);
  ph_conn_list_print(pt2pt->conn_list);
  ph_msg_list_print(pt2pt->msg_list);
  ph_routing_print(pt2pt->routing);
  ph_debug_print2("]\n\n");
}


int ph_pt2pt_get_is_disposed(ph_pt2pt_t pt2pt) {
  int n;
  ph_pthread_mutex_lock(&pt2pt->mp);
  n = pt2pt->is_disposed;
  ph_pthread_mutex_unlock(&pt2pt->mp);
  return n;
}


void ph_pt2pt_set_is_disposed(ph_pt2pt_t pt2pt, int is_disposed) {
  ph_pthread_mutex_lock(&pt2pt->mp);
  pt2pt->is_disposed = is_disposed;
  ph_pthread_mutex_unlock(&pt2pt->mp);
}


int ph_pt2pt_get_is_finalizing(ph_pt2pt_t pt2pt) {
  int n;
  ph_pthread_mutex_lock(&pt2pt->mp);
  n = pt2pt->is_finalizing;
  ph_pthread_mutex_unlock(&pt2pt->mp);
  return n;
}


void ph_pt2pt_set_is_finalizing(ph_pt2pt_t pt2pt, int is_finalizing) {
  ph_pthread_mutex_lock(&pt2pt->mp);
  pt2pt->is_finalizing = is_finalizing;
  ph_pthread_mutex_unlock(&pt2pt->mp);
}


#ifdef PH_VIEWER

int ph_pt2pt_get_is_viewing(ph_pt2pt_t pt2pt) {
  int n;
  ph_pthread_mutex_lock(&pt2pt->mp);
  n = pt2pt->is_viewing;
  ph_pthread_mutex_unlock(&pt2pt->mp);
  return n;
}


void ph_pt2pt_set_is_viewing(ph_pt2pt_t pt2pt, int is_viewing) {
  ph_pthread_mutex_lock(&pt2pt->mp);
  pt2pt->is_viewing = is_viewing;
  ph_pthread_mutex_unlock(&pt2pt->mp);
}
#endif



/* --------------------------------------------------------------------------------------------------------- */


static void initialize_select(fd_set * r_fdset_p, fd_set * w_fdset_p, fd_set * e_fdset_p) {
  if (r_fdset_p != NULL) { FD_ZERO(r_fdset_p); }
  if (w_fdset_p != NULL) { FD_ZERO(w_fdset_p); }
  if (e_fdset_p != NULL) { FD_ZERO(e_fdset_p); }
}


/* if there is at least one message that can be routed, return 1. otherwise return 0. */
static int is_ready_for_sending(ph_pt2pt_t pt2pt) {
  ph_msg_cell_t p;
   
  for (p = pt2pt->msg_list->head; p != NULL; p = p->next) {
    ph_pthread_mutex_lock(&p->self->mp);
    if (p->self->to_conn != NULL) {
      ph_pthread_mutex_unlock(&p->self->mp);
      return 1;
    }
    ph_pthread_mutex_unlock(&p->self->mp);
  }
  return 0;
}


static int sender_check_io(ph_pt2pt_t pt2pt, fd_set * w_fdset_p, fd_set * e_fdset_p) {     
  ph_msg_cell_t p;
  int n;

  initialize_select(NULL, w_fdset_p, e_fdset_p);
  
  ph_pthread_mutex_lock(&pt2pt->msg_list->mp); 
  while ((is_ready_for_sending(pt2pt) == 0) && (ph_pt2pt_get_is_disposed(pt2pt) == 0)) {
    ph_pthread_cond_wait(&pt2pt->msg_list->i_cond, &pt2pt->msg_list->mp);
  }

  if (ph_pt2pt_get_is_disposed(pt2pt)) {
    ph_pthread_mutex_unlock(&pt2pt->msg_list->mp); 
    return 0;
  }

  n = 0; /* the number of forwardable mesages */
  for (p = pt2pt->msg_list->head; p != NULL; p = p->next) {
    ph_pthread_mutex_lock(&p->self->mp);

    if ((p->self->to_conn == NULL) || (ph_conn_get_state(p->self->to_conn) == closed)) {
      // un-routable message 
      ph_pthread_mutex_unlock(&p->self->mp);
      continue;
    }
    (*p->self->to_conn->fd_set_func_p)(p->self->to_conn->attr, w_fdset_p, PH_FD_TYPE_WRITE);
    (*p->self->to_conn->fd_set_func_p)(p->self->to_conn->attr, e_fdset_p, PH_FD_TYPE_WRITE);
    ph_pthread_mutex_unlock(&p->self->mp);
  }
  ph_pthread_mutex_unlock(&pt2pt->msg_list->mp); 

  if (n) {
    struct timeval 	timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = PH_SELECT_TIMEOUT;
    select(PH_MAX_FD, NULL, w_fdset_p, e_fdset_p, &timeout);
  }
  return 1;
}


static void sender_handle_io(ph_pt2pt_t pt2pt, fd_set * w_fdset_p) {
  ph_msg_t msg;
  int n;

  ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
  while ((msg = ph_msg_list_dequeue_remote_msg(pt2pt->msg_list, w_fdset_p)) != NULL) {
    ph_conn_lock(msg->to_conn);
    ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
    
    n = ph_msg_send(msg);
    ph_conn_unlock(msg->to_conn);
    
    if (n == 1) {
      /* send successfully */
      ph_msg_dispose(msg);
    } else if (n == 0) {
      /* fail to send */
      //ph_err_msg("send failure");
      switch (msg->type) {
      case meta:
	ph_msg_dispose(msg);
	break;
      case user:
	/* re-enqueu the message */
	ph_pt2pt_handle_msg(pt2pt, msg);
	break;
      default:
	ph_err_quit("sender_handle_io() : no matching");
	break;
      } 
    }
    ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
  }
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
}


static void sender(ph_pt2pt_t pt2pt) {
  while (ph_pt2pt_get_is_disposed(pt2pt) == 0) {
    fd_set w_fdset, e_fdset;

    if (sender_check_io(pt2pt, &w_fdset, &e_fdset) == 0) {
      break;
    }
    sender_handle_io(pt2pt, &w_fdset);
  }
}


static void * sender_th(void * arg) {
  sender((ph_pt2pt_t)arg);
#ifdef __PH_WIN32__
  _endthread();
#endif
  return NULL;
}


/* --------------------------------------------------------------------------------------------------------- */


static void receiver_check_io(ph_pt2pt_t pt2pt, fd_set * r_fdset_p, fd_set * w_fdset_p, fd_set * e_fdset_p) {
  ph_conn_cell_t	p;
  struct timeval 	timeout;


  initialize_select(r_fdset_p, w_fdset_p, e_fdset_p);

  ph_pthread_mutex_lock(&pt2pt->conn_list->mp);  
  for (p = pt2pt->conn_list->head; p != NULL; p = p->next) {
    switch (ph_conn_get_state(p->self)) {
    case connecting:
      //TODO: non-blocing connect$B$N=hM}(B
      break;
    case accepting:
    case opened:
      (*p->self->fd_set_func_p)(p->self->attr, r_fdset_p, PH_FD_TYPE_READ);
      (*p->self->fd_set_func_p)(p->self->attr, e_fdset_p, PH_FD_TYPE_ERROR);
      break;
    case closed:
      break;
    default:
      ph_err_quit("receive_check_io() : no matching");
      break;
    }
  }
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
  

  timeout.tv_sec = 0;
  timeout.tv_usec = PH_SELECT_TIMEOUT;
  select(PH_MAX_FD, r_fdset_p, w_fdset_p, e_fdset_p, &timeout);
}


static void receiver_handle_io(ph_pt2pt_t pt2pt, ph_conn_t conn, fd_set * r_fdset_p, fd_set * w_fdset_p, fd_set * e_fdset_p) {
  switch (ph_conn_get_state(conn)) {
  case connecting:
    ph_pt2pt_handle_connecting(pt2pt, conn, r_fdset_p, w_fdset_p);
    break;
  case accepting:
    ph_pt2pt_handle_accepting(pt2pt, conn, r_fdset_p);
    break;
  case opened:
    ph_pt2pt_handle_opened(pt2pt, conn, r_fdset_p);
    break;
  case closed:
    break;
  default:
    ph_err_quit("receiver_handle_io() : no matching");
    break;
  }
  ph_pt2pt_handle_error(pt2pt, conn, e_fdset_p);
}


static void receiver(ph_pt2pt_t pt2pt) {
  while (ph_pt2pt_get_is_disposed(pt2pt) == 0) {
    fd_set r_fdset, w_fdset, e_fdset;
    ph_conn_cell_t p;

    receiver_check_io(pt2pt, &r_fdset, &w_fdset, &e_fdset);

    for (p = ph_conn_list_init_loop(pt2pt->conn_list); p != NULL; p = ph_conn_list_iterate(pt2pt->conn_list, p)) {
      receiver_handle_io(pt2pt, p->self, &r_fdset, &w_fdset, &e_fdset);
    }
  }
}


static void * receiver_th(void * arg) {
  receiver((ph_pt2pt_t)arg);
#ifdef __PH_WIN32__
  _endthread();
#endif
  return NULL;
}


/* --------------------------------------------------------------------------------------------------------- */

#ifdef		PH_ROUTING_ENABLE_DOMINATING


/* compare connection (p,q) and (pp,qq) */
static int compare(ph_vp_set_t p0, ph_vp_set_t q0, int addr_type0, 
	    ph_vp_set_t p1, ph_vp_set_t q1, int addr_type1) {
  ph_vp_t r0, r1;
  
  if ((addr_type0 == tcp) && (addr_type1 == ssh)) {
    return 0;
  }

  if ((addr_type0 == ssh) && (addr_type1 == tcp)) {
    return 1;
  }

  if (ph_vp_set_is_equal(p0, p1)) {
    return (ph_vp_set_get_max_elem(q0) > ph_vp_set_get_max_elem(q1));
  }

  if (ph_vp_set_is_equal(q0, q1)) {
    return (ph_vp_set_get_max_elem(p0) > ph_vp_set_get_max_elem(p1));
  }

  if (ph_vp_set_is_equal(p0, q1)) {
    return (ph_vp_set_get_max_elem(q0) > ph_vp_set_get_max_elem(p1));
  }

  if (ph_vp_set_is_equal(q0, p1)) {
    return (ph_vp_set_get_max_elem(p0) > ph_vp_set_get_max_elem(q1));
  }
  
  r0 = ph_max(ph_vp_set_get_max_elem(p0), ph_vp_set_get_max_elem(q0));
  r1 = ph_max(ph_vp_set_get_max_elem(p1), ph_vp_set_get_max_elem(q1));

  return (r0 > r1);
}



/* return 1 if connection between I and p must be deleted.
   (if I and p and q form a triangle and the connection between me and p is the highest) */
int check_triangle(ph_vp_set_t local_vp_set, ph_routing_neigh_t p, ph_addr_type_t p_type, ph_routing_neigh_t q, ph_addr_type_t q_type) {
  int type; /* addr_type of the connection between p and q */
  
  /* for tcp and ssh */
  for (type = 0; type < 2; type++) { 
    /* if p and q are adjacent, I and p and q form a triangle */
    if (ph_vp_set_is_intersected(p->local, q->p_neigh[type])) { /* are p and q connected with type? */
      
      /* if (I,p) < (I,q) and (I,p) < (p,q) then (I,p) must be deleted */
      if ((compare(local_vp_set, p->local, p_type, 
		   local_vp_set, q->local, q_type)) &&

	  (compare(local_vp_set, p->local, p_type, 
		   p->local, q->local, type))) {
#if 0	
	{
	  
	  ph_vp_range_cell_t vps;
	  
	  fprintf(stderr, "local=");
	  for (vps = local_vp_set->head; vps != NULL; vps = vps->next) {
	    fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", vps->self->lower, vps->self->upper, ((vps->next != NULL) ? "," : ""));
	  }
	 fprintf(stderr, "]   p=");
	 for (vps = p->local->head; vps != NULL; vps = vps->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", vps->self->lower, vps->self->upper, ((vps->next != NULL) ? "," : ""));
	 }
	 fprintf(stderr, "]   q=");
	 for (vps = q->local->head; vps != NULL; vps = vps->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", vps->self->lower, vps->self->upper, ((vps->next != NULL) ? "," : ""));
	 }
	 fprintf(stderr, "]\n\n");
	}
#endif

	return 1;
      }
    }
  }
    
  return 0;
}


/* return 0 if I and p and q and p's neighbor form a square and the connection between me and p is the highest */
int check_square(ph_vp_set_t local_vp_set, ph_routing_neigh_t p, ph_addr_type_t p_type, ph_routing_neigh_t q, ph_addr_type_t q_type) {
  int type0, type1; /* addr_type of the connection between p and q */

   
  /* for tcp and ssh */
  for (type0 = 0; type0 < 2; type0++) { 
    for (type1 = 0; type1 < 2; type1++) { 
	ph_vp_set_t tmp, vp_set;


      /* get intersection of p->neigh[type0], q->neigh[type1] to determine the max vp of p and q's shared neighbor */
	tmp = ph_vp_set_make_intersection(p->p_neigh[type0], q->p_neigh[type1]);
	vp_set = ph_vp_set_make_diff(tmp, local_vp_set);
	ph_vp_set_dispose(tmp);

      /* if (p and q are adjacent) and (p's neighbor and q are adjacent), they form a square */
      if (ph_vp_set_is_empty(vp_set) == 0)  { /* are p's neighbor and q connected with type? */
#if 0
	{
	  ph_vp_range_cell_t vps;
	  	  
	  fprintf(stderr, "local=");
	  for (vps = local_vp_set->head; vps != NULL; vps = vps->next) {
	    fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", vps->self->lower, vps->self->upper, ((vps->next != NULL) ? "," : ""));
	  }
	 fprintf(stderr, "]   p=");
	 for (vps = p->local->head; vps != NULL; vps = vps->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", vps->self->lower, vps->self->upper, ((vps->next != NULL) ? "," : ""));
	 }
	 fprintf(stderr, "]   q=");
	 for (vps = q->local->head; vps != NULL; vps = vps->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", vps->self->lower, vps->self->upper, ((vps->next != NULL) ? "," : ""));
	 }
	 fprintf(stderr, "]   vp_set=");
	 for (vps = vp_set->head; vps != NULL; vps = vps->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", vps->self->lower, vps->self->upper, ((vps->next != NULL) ? "," : ""));
	 }
	 fprintf(stderr, "]\n\n");
	}
#endif

	/* if (I,p) < (I,q) and (I,p) < (p,q) and (I,p) < (p's neighbor,q), then (I,p) must be deleted */
	if ((compare(local_vp_set, p->local, p_type, 
		     local_vp_set, q->local, q_type)) &&

	    (compare(local_vp_set, p->local, p_type, 
		     vp_set, p->local, type0)) &&

	    (compare(local_vp_set, p->local, p_type, 
		     vp_set, q->local, type1))) {
	  ph_vp_set_dispose(vp_set);
	  return 1;
	}
      }
      ph_vp_set_dispose(vp_set);

    }
  }
  return 0;
}



static int is_redundant(ph_pt2pt_t pt2pt, ph_conn_t conn) {
  ph_routing_t routing = pt2pt->routing;
  ph_routing_neigh_t p; /* p->conn == conn */
  ph_routing_neigh_cell_t q;
  ph_vp_set_t local_vp_set;

  /* only delete ssh connection */
  if (ph_conn_get_addr_type(conn) != ssh) {
    return 0;
  }


  if ((p = ph_routing_neigh_list_find(routing->neigh_list, &ph_routing_neigh_is_same_conn, conn)) == NULL) {
    /* neighbor info has net yet been reached. */
    return 0;
  }

  local_vp_set = ph_vp_set_sync_dup(pt2pt->vp_set);
  
  /* check all triangles and squares that include me and p */
  for (q = routing->neigh_list->head; q != NULL; q = q->next) {
    ph_addr_type_t p_type, q_type;
    
    /* p and q are my neighbors */
    assert(p->conn != NULL);
    assert((q->self->conn != NULL) || (q->self->addr != NULL));
    p_type = (p->conn != NULL) ? ph_conn_get_addr_type(p->conn) : p->addr->type;
    q_type = (q->self->conn != NULL) ? ph_conn_get_addr_type(q->self->conn) : q->self->addr->type;
    
    /* do nothing if p and q point the same node */
    if (ph_vp_set_is_equal(q->self->local, p->local)) {
      continue;
    }

    /* check if I and p and q form a triangle */
    if (check_triangle(local_vp_set, p, p_type, q->self, q_type)) {
      ph_vp_set_dispose(local_vp_set);
      return 1;
    }

    /* check if I and p p's neighbor and q form a square */
    if (check_square(local_vp_set, p, p_type, q->self, q_type)) {
      ph_vp_set_dispose(local_vp_set);
      return 1;
    }
  }

  ph_vp_set_dispose(local_vp_set);
  return 0;
}

static void connector_check(ph_pt2pt_t pt2pt) {
  ph_conn_cell_t p;
  ph_vp_set_t vps[2], local_vps;
  ph_routing_neigh_cell_t q;
  struct timeval t;

  
  /* TODO: $B8E$$>pJs$r$$$D:o=|$9$k$+(B
                $B"#(B $B0lDj;~4VAj<j$+$i<+J,$N@8@.$G$-$F$$$J$$(B addr $B$X$N(B possible neigh $B$,$"$C$?$i$=$l$r:o=|$9$k(B
                $B"#(B $B0lDj;~4VAj<j$+$iD%$i$l$F$$$J$$(B possible neigh $B$,$"$C$?$i$=$l$r:o=|$9$k(B
           $B$$$D?7$7$$>pJs$r3MF@$7$h$&$H$9$k$+(B
                   $B$$$D%3%M%/%7%g%s$rD%$j$K9T$/$+(B
                     - $B<+J,$N(B neigh $B$N>pJs$,99?7$5$l$?$H$-(B,
                     - $BDj4|E*$K(B
  */


  local_vps = ph_vp_set_sync_dup(pt2pt->vp_set);

  ph_pthread_mutex_lock(&pt2pt->routing->mp);
  ph_pthread_mutex_lock(&pt2pt->conn_list->mp);

  /* delete obsolete information */
  /*
  gettimeofday(&t, NULL);
  q = pt2pt->routing->neigh_list->head;
  while (q != NULL) { 
    double passed;
    ph_routing_neigh_cell_t next = q->next;
    
    passed = (double)(t.tv_sec - q->self->time.tv_sec) + ((double)(t.tv_usec - q->self->time.tv_usec))/1000000.0;

    if (passed > 30.0) { 
      ph_routing_neigh_list_delete(pt2pt->routing->neigh_list, q->self);
    }
    q = next;
  }
  */


  vps[tcp] = ph_vp_set_make_empty();
  vps[ssh] = ph_vp_set_make_empty();
  for (q = pt2pt->routing->neigh_list->head; q != NULL; q = q->next) {
    ph_vp_set_t tmp;

    if (q->self->conn == NULL) {
      continue;
    }
    switch (ph_conn_get_addr_type(q->self->conn)) {
    case tcp:
    case unknown:
      tmp = ph_vp_set_make_union(vps[tcp], q->self->local);
      ph_vp_set_dispose(vps[tcp]);
      vps[tcp] = tmp;
      break;
    case ssh:
      tmp = ph_vp_set_make_union(vps[ssh], q->self->local);
      ph_vp_set_dispose(vps[ssh]);
      vps[ssh] = tmp;
      break;
    default:
      ph_err_quit("connector_check() : no matching");
      break;
    }
  }
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
  ph_pthread_mutex_unlock(&pt2pt->routing->mp);

  /* $B>pJs$,99?7$5$l$?$+(B, $B?7$7$/@8@.$7$?%3%M%/%7%g%s$KBP$7$F%a%C%;!<%8$rAw?.$9$k(B */

  /* send my vps and all possible neighbor's vps */
  for (p = ph_conn_list_init_loop(pt2pt->conn_list); p != NULL; p = ph_conn_list_iterate(pt2pt->conn_list, p)) {
    if (ph_conn_get_state(p->self) != opened) {
      continue;
    }
    ph_pt2pt_handle_msg(pt2pt, ph_pt2pt_create_meta_msg(pt2pt, p->self, conn_type, ph_conn_get_addr_type(p->self), local_vps, vps[tcp], vps[ssh]));
  }

  ph_vp_set_dispose(local_vps);
  ph_vp_set_dispose(vps[tcp]);
  ph_vp_set_dispose(vps[ssh]);
}


static int is_useless(ph_pt2pt_t pt2pt, ph_conn_t conn) {
  int n;
  int is_repaired = 0;

  assert(conn != NULL);
   

  if (ph_conn_get_state(conn) == opened) { 

    if (is_redundant(pt2pt, conn)) { /* $B%a%C%;!<%8$NAw<u?.$,=*$o$C$?8e$K7W;;$7$F$$$$(B */
      //ph_conn_set_state(conn, closed);
      ph_conn_set_is_permanent(conn, 0);
      //ph_routing_event_list_enqueue(pt2pt->routing->event_list, ph_routing_event_new(fail, conn, ph_routing_entry_list_new(0)));

      ph_pthread_mutex_lock(&conn->mp);
      //fprintf(stderr, "redundant %s:%s\n", conn->addr->host, conn->addr->service);
      ph_pthread_mutex_unlock(&conn->mp);
      
    } else if (ph_conn_get_is_permanent(conn) == 0) {
      /* if it is a newly accepted connection, broadcast routing table update messages */
      is_repaired = 1;
      ph_conn_set_is_permanent(conn, 1);
    }
  }
  
  
  if (ph_conn_get_state(conn) == closed) {
    int flag = 0;
    ph_routing_entry_cell_t q;

    flag = 1;
    /*
    for (q = pt2pt->routing->entry_list->head; q != NULL; q = q->next) {
      if (q->self->conn == conn) { 
	flag = 1;
      }
    }
    */

    if (flag) {
      //ph_routing_event_list_enqueue(pt2pt->routing->event_list, ph_routing_event_new(fail, conn, ph_routing_entry_list_new(0)));
    }
    ph_conn_set_is_permanent(conn, 0);
  }
  
  if (is_repaired) {
    //ph_routing_event_list_enqueue(pt2pt->routing->event_list, ph_routing_event_new(repair, conn, NULL)); 
  }
  
  
  ph_pthread_mutex_lock(&conn->mp);
  n = ((conn->state == closed) && (conn->nref == 0));
  ph_pthread_mutex_unlock(&conn->mp);
  
  return n;
}


static void connector_cleanup(ph_pt2pt_t pt2pt) {
  ph_conn_cell_t p;
  
  ph_pthread_mutex_lock(&pt2pt->routing->mp);
  ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
  p = pt2pt->conn_list->head;
  while (p != NULL) {
    ph_conn_cell_t next = p->next;

    //fprintf(stderr, "[%s:%s] closed=%d, nref=%d\n", p->self->addr->host, p->self->addr->service, p->self->state == closed, p->self->nref);

    if (is_useless(pt2pt, p->self)) {
      ph_routing_neigh_t neigh;

      ph_msg_list_clear_conn(pt2pt->msg_list, p->self); // TODO
      
      /* $B$b$&$9$3$7$-$l$$$K=q$/$Y$-(B */
      while ((neigh = ph_routing_neigh_list_find(pt2pt->routing->neigh_list, &ph_routing_neigh_is_same_conn, p->self)) != NULL) {
	int i;	
	
        //ph_routing_neigh_list_delete(pt2pt->routing->neigh_list, neigh);
	
	neigh->addr = ph_addr_dup(neigh->conn->addr);
	neigh->conn = NULL;
	for (i = 0; i < 2; i++) {
	  ph_vp_set_dispose(neigh->neigh[i]);
	  neigh->neigh[i] = ph_vp_set_make_empty();
        }
      }


      /* $B%k!<%F%#%s%0I=$N(Bentry$BCf$K$OB8:_$7$J$$$3$H$r<($9(B */
      {
	ph_routing_entry_cell_t q;

	for (q = pt2pt->routing->entry_list->head; q != NULL; q = q->next) {
	  if (q->self->conn == p->self) {
	    ph_err_quit("(q->self->conn != p->self (ref=%d) closed?=%d", p->self->nref, p->self->state == closed);
	  }
	}
      }

      //fprintf(stderr, "delete [%s:%s]\n", p->self->addr->host, p->self->addr->service);
      ph_conn_list_delete_opt(pt2pt->conn_list, p->self, 0);
      
    }
    p = next;
  }
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
  ph_pthread_mutex_unlock(&pt2pt->routing->mp);
}



static void connector_establish(ph_pt2pt_t pt2pt) {
  ph_addr_cell_t p;
  int type;
  struct timeval t[2];

  gettimeofday(&t[0], NULL);

  for (type = 0; type < 2; type++) { 

    for (p = ph_addr_list_init_loop(pt2pt->addr_list); p != NULL; p = ph_addr_list_iterate(pt2pt->addr_list, p)) {
      /* first establish regular tcp connections, and then establish SSH connections */
      if (p->self->type != type) {
	continue;
      }

      if (ph_conn_list_find_same_addr(pt2pt->conn_list, p->self) == NULL) {
	int n;
	/* $B%3%M%/%7%g%s$rD%$C$F$$$J$$$N$O(B,
	   [1] connect() $B$,@.8y$7$J$$(B
	   [2] MST$B$H$7$FA*Br$5$l$J$+$C$?%3%M%/%7%g%s(B
	   $B$N(B2$BDL$j$N2DG=@-$,$"$k!%(B
	   
	   neighbor $B$N>pJs$,99?7$5$l$kEY$K%a%C%;!<%8$rAw?.$9$kI,MW$,$"$k!)(B
	   
	*/
	//fprintf(stderr, "connect %s:%s\n", p->self->host, p->self->service);
	
	
	ph_pthread_mutex_lock(&pt2pt->routing->mp);
	n = (ph_routing_neigh_list_find_same_addr(pt2pt->routing->neigh_list, p->self) == NULL);
	ph_pthread_mutex_unlock(&pt2pt->routing->mp);
	
	ph_pt2pt_handle_connect(pt2pt, p->self);
	
	/*
	if (n) {
	  ph_pt2pt_handle_connect(pt2pt, p->self);
	}
	*/

	
	/* $B@\B3$K<:GT$7$?;~E@$G!$$=$N(Bneighbor $B$N>pJs$r:o=|$9$k(B */
	
      }else {
	//fprintf(stderr, "already connect %s:%s\n", p->self->host, p->self->service);
      }
      
      {
	double passed;
	gettimeofday(&t[1], NULL);
	passed = (double)(t[1].tv_sec - t[0].tv_sec) + ((double)(t[1].tv_usec - t[0].tv_usec))/1000000.0;
	
	if (passed > 5) {

	  gettimeofday(&t[0], NULL);

	  //connector_check(pt2pt);
	  //connector_cleanup(pt2pt);
	  
	}
      }

      

      if (ph_pt2pt_get_is_disposed(pt2pt)) {
	break;
      }
    }
  }
}

static void my_check(ph_pt2pt_t pt2pt, struct timeval t) {
  double passed;
  char hostname[8192];
  struct timeval tt;
  

  memset(hostname, 0, 8192);
  gethostname(hostname, 8192);
  
  gettimeofday(&tt, NULL);
  passed = (double)(tt.tv_sec - t.tv_sec)
    + ((double)(tt.tv_usec - t.tv_usec))/1000000.0;
  
  {
    ph_conn_cell_t p;
    int flag;

    if (strncmp(hostname, "csc", 3) == 0) {
      int no;
      no = atoi(hostname +3);
      
      if (no == 0) {
	flag = 0;
      } else {
	flag = 1;
      }
    } else if (strncmp(hostname, "istsun", 6) == 0) {
      flag = 0;
    } else if (strncmp(hostname, "uguisu", 6) == 0) { 
      int no;
      no = atoi(hostname +6);
      if (no == 0) {
	flag = 0;
      } else {
	flag = 1;
      }
    }

    ph_pthread_mutex_lock(&pt2pt->routing->mp);
    ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
    for (p = pt2pt->conn_list->head; p != NULL; p = p->next) { 
      ph_pthread_mutex_lock(&p->self->mp);
      //if ((p->self->addr->type == ssh) && (p->self->state == opened)) {
      if ((p->self->addr->type == ssh) && (p->self->is_permanent == 1)) {
	
	if (strncmp(hostname, "csc", 3) == 0) {
	  int no;

	  no = atoi(hostname +3);
	  
	  if (no == 0) {
	    if (strcmp(p->self->addr->host, "133.11.36.8") == 0) {
	      flag = 1;
	    } else {
	      fprintf(stderr, "%s\n", p->self->addr->host);
	      flag = 0;
	      ph_pthread_mutex_unlock(&p->self->mp);
	      break;
	    }
	  } else {
	    flag = 0;
	    ph_pthread_mutex_unlock(&p->self->mp);
	    break;
	  }

	} else if (strncmp(hostname, "istsun", 6) == 0) {
	  
	  if (strcmp(p->self->addr->host, "xxxx") == 0) {
	    flag = 1;
	  } else {
	    fprintf(stderr, "%s\n", p->self->addr->host);
	    flag = 0;
	    ph_pthread_mutex_unlock(&p->self->mp);
	    break;
	  }

	} else if (strncmp(hostname, "uguisu", 6) == 0) { 
	  int no;

	  no = atoi(hostname +6);
	  
	  if (no == 0) {
	    if ((strcmp(p->self->addr->host, "133.11.23.16") == 0) || 
		(strcmp(p->self->addr->host, "xxxx") == 0))  {
	      flag = 1;
	    } else {
	      flag = 0;
	      ph_pthread_mutex_unlock(&p->self->mp);
	      break;
	    }
	  } else {
	    fprintf(stderr, "no=%d, host=%s\n", no, p->self->addr->host);
	    flag = 0;
	    ph_pthread_mutex_unlock(&p->self->mp);
	    break;
	  }
	}

      }
      ph_pthread_mutex_unlock(&p->self->mp);
    }
    ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
    ph_pthread_mutex_unlock(&pt2pt->routing->mp);

    if (flag) {
      if (pt2pt->checker == 0) {
	fprintf(stderr, "%s : success (%f)\n", hostname, passed);
      }
      pt2pt->checker = 1;
    } else {
      fprintf(stderr, "%s : fail    (%f)\n", hostname, passed);
      pt2pt->checker = 0;
    }
    
    if (passed > 60.0) {
      //fprintf(stderr, "60 seconds has passed.\n");
      //connector_cleanup(pt2pt);
    } 
  }
}
  

static void connector(ph_pt2pt_t pt2pt) {
  struct timeval t[2];

  gettimeofday(&t[0], NULL);


  while (ph_pt2pt_get_is_disposed(pt2pt) == 0) {
    connector_establish(pt2pt);
    connector_check(pt2pt);
    

#if 0
    {
      ph_routing_t routing = pt2pt->routing;
      ph_routing_neigh_cell_t p;
      ph_pthread_mutex_lock(&routing->mp);
      for (p = routing->neigh_list->head; p != NULL; p = p->next) { 
	 ph_vp_range_cell_t q;
	 
	 fprintf(stderr, "local=");
	 for (q = p->self->local->head; q != NULL; q = q->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", q->self->lower, q->self->upper, ((q->next != NULL) ? "," : ""));
	 }
	 /*
	 fprintf(stderr, "]   neigh[0]=");
	 for (q = p->self->neigh[0]->head; q != NULL; q = q->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", q->self->lower, q->self->upper, ((q->next != NULL) ? "," : ""));
	 }
	 fprintf(stderr, "]   neigh[1]=");
	 for (q = p->self->neigh[1]->head; q != NULL; q = q->next) {
	 fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", q->self->lower, q->self->upper, ((q->next != NULL) ? "," : ""));
	 }
	 */

	 fprintf(stderr, "] p_neigh[0]=");
	 for (q = p->self->p_neigh[0]->head; q != NULL; q = q->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", q->self->lower, q->self->upper, ((q->next != NULL) ? "," : ""));
	 }
	 fprintf(stderr, "] p_neigh[1]=");
	 for (q = p->self->p_neigh[1]->head; q != NULL; q = q->next) {
	   fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s", q->self->lower, q->self->upper, ((q->next != NULL) ? "," : ""));
	 }

	 fprintf(stderr, "]\n");
      }
      fprintf(stderr, "\n");
      ph_pthread_mutex_unlock(&routing->mp);
    }
#endif

    sleep(3);

    {
     
      connector_cleanup(pt2pt);

      my_check(pt2pt, t[0]);

    }
  }
}


#else 



static void connector_establish_connections(ph_pt2pt_t pt2pt) {
  ph_addr_cell_t p;

  for (p = ph_addr_list_init_loop(pt2pt->addr_list); (p != NULL) && (ph_conn_list_get_nelm(pt2pt->conn_list) < PH_MAX_CONNECTION); p = ph_addr_list_iterate(pt2pt->addr_list, p)) {
    if (ph_conn_list_find_same_addr(pt2pt->conn_list, p->self) == NULL) {
      ph_pt2pt_handle_connect(pt2pt, p->self);
    }

    if (ph_pt2pt_get_is_disposed(pt2pt)) {
      break;
    }
  }
}

static int is_useless(ph_conn_t conn) {
  int n;

  assert(conn != NULL);

  ph_pthread_mutex_lock(&conn->mp);
  n = ((conn->state == closed) && (conn->nref == 0));
  ph_pthread_mutex_unlock(&conn->mp);

  return n;
}

static void connector_cleanup_connections(ph_pt2pt_t pt2pt) {
  ph_conn_cell_t p;

  ph_pthread_mutex_lock(&pt2pt->conn_list->mp);
  p = pt2pt->conn_list->head;
  while (p != NULL) {
    ph_conn_cell_t next;

    next = p->next;
    if (is_useless(p->self)) {
      ph_msg_list_clear_conn(pt2pt->msg_list, p->self); // TODO
      ph_conn_list_delete_opt(pt2pt->conn_list, p->self, 0);
    }
    p = next;
  }
  ph_pthread_mutex_unlock(&pt2pt->conn_list->mp);
}


static void connector(ph_pt2pt_t pt2pt) {
  while (ph_pt2pt_get_is_disposed(pt2pt) == 0) {
    connector_establish_connections(pt2pt);
    connector_cleanup_connections(pt2pt);
#ifndef __PH_WIN32__
    //usleep(PH_CONNECTOR_INTERVAL);
    sleep(1);
#else
    Sleep(PH_CONNECTOR_INTERVAL);
#endif
  }
}

#endif

static void * connector_th(void * arg) {
  connector((ph_pt2pt_t)arg);
#ifdef __PH_WIN32__
  _endthread();
#endif
  return NULL;
}


/* --------------------------------------------------------------------------------------------------------- */


static void router_handle_event_list(ph_pt2pt_t pt2pt, int * is_inc_update_p, double * inc_update_time_p) {
  ph_routing_event_t event;
  
  while ((event = ph_routing_event_list_dequeue_opt(pt2pt->routing->event_list, 0, NULL, NULL, NULL)) != NULL) {
    ph_routing_handle_event(pt2pt->routing, event, is_inc_update_p, inc_update_time_p);
    ph_routing_event_dispose(event);
  }
  ph_routing_route_msg_list(pt2pt->routing, pt2pt->msg_list);
}


/* TODO: P(u,v) $B$N@.$jN)$?$J$$%3%M%/%7%g%s$G$O%a%C%;!<%8$rAw?.$7$J$$(B */
static void router_send_update_message(ph_pt2pt_t pt2pt, double full_update_time, int is_inc_update, double inc_update_time) {
  ph_conn_cell_t p;
  char * full_entries, * inc_entries = NULL;
  size_t full_len, inc_len;

  
  /* create bodies of messages */
  full_entries = ph_routing_create_msg(pt2pt->routing, &full_len, 1, ph_pt2pt_get_is_finalizing(pt2pt));

  if ((is_inc_update) && (inc_update_time <= 0.0)) {
    inc_entries = ph_routing_create_msg(pt2pt->routing, &inc_len, 0, ph_pt2pt_get_is_finalizing(pt2pt));
  }

  for (p = ph_conn_list_init_loop(pt2pt->conn_list); p != NULL; p = ph_conn_list_iterate(pt2pt->conn_list, p)) {
#ifdef PH_ROUTING_ENABLE_DOMINATING
    if ((ph_conn_get_state(p->self) != opened) || (ph_conn_get_is_permanent(p->self) == 0)) {
      //if (ph_conn_get_state(p->self) != opened) {
#else  
   if (ph_conn_get_state(p->self) != opened) {
#endif
      continue;
    }

    if (full_update_time <= 0.0) {
      /*
      fprintf(stderr, "send full_update message to (%s:%s)\n", 
	      (((p->self->addr != NULL) && (p->self->addr->host != NULL)) ? p->self->addr->host : "(null)"), 
	      (((p->self->addr != NULL) && (p->self->addr->service != NULL)) ? p->self->addr->service : "(null)"));
      */
      ph_pt2pt_handle_msg(pt2pt, 
			  ph_pt2pt_create_meta_msg(pt2pt, p->self, routing, full_entries, full_len));
    } else if (p->self->is_repaired) {
      /*
      fprintf(stderr, "send full_update message (repair) to (%s:%s)\n", 
	      (((p->self->addr != NULL) && (p->self->addr->host != NULL)) ? p->self->addr->host : "(null)"), 
	      (((p->self->addr != NULL) && (p->self->addr->service != NULL)) ? p->self->addr->service : "(null)"));
      */

    ph_pt2pt_handle_msg(pt2pt, 
			     ph_pt2pt_create_meta_msg(pt2pt, p->self, routing, full_entries, full_len));
      p->self->is_repaired = 0;
    } else if ((is_inc_update) && (inc_update_time <= 0.0) && (inc_entries != NULL)) {
      /*
      fprintf(stderr, "send inc_update message to (%s:%s)\n", 
	      (((p->self->addr != NULL) && (p->self->addr->host != NULL)) ? p->self->addr->host : "(null)"), 
	      (((p->self->addr != NULL) && (p->self->addr->service != NULL)) ? p->self->addr->service : "(null)"));
      */

    ph_pt2pt_handle_msg(pt2pt, 
			ph_pt2pt_create_meta_msg(pt2pt, p->self, routing, inc_entries, inc_len));      
    }
  }
  
  ph_free(full_entries);
  if (inc_entries != NULL) { ph_free(inc_entries); }
}


static void router(ph_pt2pt_t pt2pt) { 
  double full_update_time = (double)PH_ROUTER_FULL_UPDATE_INTERVAL;
  double inc_update_time = 0.0;
  int is_inc_update = 0;


  while (ph_pt2pt_get_is_disposed(pt2pt) == 0) {
    {
      double t;

      t = ph_routing_wait_event_list(pt2pt->routing, (is_inc_update) ? ph_min(inc_update_time, full_update_time) : full_update_time);
      full_update_time -= t;

      if (is_inc_update) { inc_update_time -= t; }

      if (ph_pt2pt_get_is_disposed(pt2pt)) {
	break;
      }

      router_handle_event_list(pt2pt, &is_inc_update, &inc_update_time);
    }

    if (full_update_time <= 0.0) {
      ph_routing_increment_sequences(pt2pt->routing);
    }

    router_send_update_message(pt2pt, full_update_time, is_inc_update, inc_update_time);
 
    if (full_update_time <= 0.0) {
      full_update_time = (double)PH_ROUTER_FULL_UPDATE_INTERVAL;
    }
    if ((is_inc_update) && (inc_update_time <= 0.0)) {
      is_inc_update = 0;
    }

    ph_pt2pt_print(pt2pt); /* print debug message */

  }
}


static void * router_th(void * arg) {
  router((ph_pt2pt_t)arg);
#ifdef __PH_WIN32__
  _endthread();
#endif
  return NULL;
}


/* --------------------------------------------------------------------------------------------------------- */


#ifdef PH_VIEWER

static void viewer_connect(ph_pt2pt_t pt2pt) {
  if (pt2pt->viewer_conn == NULL) {
    pt2pt->viewer_conn = ph_conn_connect(pt2pt->viewer_addr);
  }
}


static void viewer_send(ph_pt2pt_t pt2pt) {
  ph_msg_t msg;
  fd_set w_fdset, e_fdset;


  if (pt2pt->viewer_conn == NULL) {
    return;
  }

  initialize_select(NULL, &w_fdset, &e_fdset);
  (*pt2pt->viewer_conn->fd_set_func_p)(pt2pt->viewer_conn->attr, &w_fdset, PH_FD_TYPE_WRITE);
  (*pt2pt->viewer_conn->fd_set_func_p)(pt2pt->viewer_conn->attr, &e_fdset, PH_FD_TYPE_WRITE);
  
  select(PH_MAX_FD, NULL, &w_fdset, &e_fdset, NULL);
  
  if ((*pt2pt->viewer_conn->fd_isset_func_p)(pt2pt->viewer_conn->attr, &e_fdset, PH_FD_TYPE_WRITE)) {
    ph_conn_dispose(pt2pt->viewer_conn);
    pt2pt->viewer_conn = NULL;
    return;
  }    
  
  if ((*pt2pt->viewer_conn->fd_isset_func_p)(pt2pt->viewer_conn->attr, &w_fdset, PH_FD_TYPE_WRITE)) {
    msg = ph_pt2pt_create_meta_msg(pt2pt, pt2pt->viewer_conn, view);
    if (ph_msg_send(msg) == 0) {
      /* fail to send; close the connectoin */
      ph_conn_dispose(pt2pt->viewer_conn);
      pt2pt->viewer_conn = NULL;
    }
    ph_msg_dispose(msg);
  }
}


static void viewer(ph_pt2pt_t pt2pt) { 
  while (ph_pt2pt_get_is_disposed(pt2pt) == 0) {
    viewer_connect(pt2pt);
    viewer_send(pt2pt);
    assert(pt2pt->viewer_update_interval >= 0);
    sleep(pt2pt->viewer_update_interval);
  }
}


static void * viewer_th(void * arg) {
  viewer((ph_pt2pt_t)arg);
#ifdef __PH_WIN32__
  _endthread();
#endif
  return NULL;
}


void ph_pt2pt_viewer_run(ph_pt2pt_t pt2pt, int viewer_update_interval) {
#ifndef __PH_WIN32__
  pthread_attr_t attr;
#endif

  ph_pt2pt_set_is_viewing(pt2pt, 1);
  pt2pt->viewer_update_interval = viewer_update_interval;

#ifndef __PH_WIN32__
  ph_pthread_attr_init(&attr);
  ph_pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
  ph_pthread_create(&pt2pt->viewer_tid, &attr, &viewer_th, (void *)pt2pt);
#else
  ph_begin_thread(&pt2pt->viewer_tid, &sender_th, (void*)pt2pt);
#endif
}

#endif

/* --------------------------------------------------------------------------------------------------------- */


static void server_spawn_threads(ph_pt2pt_t pt2pt) {
#ifndef __PH_WIN32__
  pthread_attr_t attr;
 
  ph_pthread_attr_init(&attr);
  ph_pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
  ph_pthread_create(&pt2pt->sender_tid, &attr, &sender_th, (void *)pt2pt);
  ph_pthread_create(&pt2pt->receiver_tid, &attr, &receiver_th, (void *)pt2pt);
  ph_pthread_create(&pt2pt->connector_tid, &attr, &connector_th, (void *)pt2pt);
  ph_pthread_create(&pt2pt->router_tid, &attr, &router_th, (void *)pt2pt);
  ph_pthread_attr_destroy(&attr);
#else
  ph_begin_thread(&pt2pt->sender_tid, &sender_th, (void*)pt2pt);
  ph_begin_thread(&pt2pt->receiver_tid, &receiver_th, (void*)pt2pt);
  ph_begin_thread(&pt2pt->connector_tid, &connector_th, (void*)pt2pt);
  ph_begin_thread(&pt2pt->router_tid, &router_th, (void*)pt2pt);
#endif
}


static void server_join_threads(ph_pt2pt_t pt2pt) {
  void 		* status;

#ifdef PH_VIEWER
  if (ph_pt2pt_get_is_viewing(pt2pt)) {
    ph_pthread_join(pt2pt->viewer_tid, &status);
  }
#endif

  ph_pthread_mutex_lock(&pt2pt->msg_list->mp);
  ph_pthread_cond_broadcast(&pt2pt->msg_list->i_cond);
  ph_pthread_mutex_unlock(&pt2pt->msg_list->mp);

  ph_pthread_join(pt2pt->sender_tid, &status);
  ph_pthread_join(pt2pt->receiver_tid, &status);
  ph_pthread_join(pt2pt->connector_tid, &status); 

  ph_routing_event_list_enqueue(pt2pt->routing->event_list, ph_routing_event_new(finalize, NULL, NULL)); // dummy
  ph_pthread_join(pt2pt->router_tid, &status); 
}


void ph_pt2pt_run(ph_pt2pt_t pt2pt) {
  ph_conn_list_insert(pt2pt->conn_list, ph_conn_listen(pt2pt->available_ports, &pt2pt->listen_port)); 
  server_spawn_threads(pt2pt);
}

void ph_pt2pt_stop(ph_pt2pt_t pt2pt) {
  server_join_threads(pt2pt);
}
