/**************************************
 * $Author: kaneda $
 * $Date: 2003/02/12 12:45:20 $
 **************************************/

/* routing.c */

#include "ph_routing.h"


/* Public Function Prototypes */


ph_routing_entry_t ph_routing_entry_new(ph_vp_t lower, ph_vp_t upper, ph_conn_t conn, int metric, int sequence, int is_modified, double st_time[], ph_addr_type_t addr_type) {
  ph_routing_entry_t p;
  int i;

  p = (ph_routing_entry_t)ph_malloc(sizeof(struct ph_routing_entry));

  p->dest = ph_vp_range_new(lower, upper);
  p->conn = conn;
  p->addr_type = addr_type;
  p->metric = metric;
  p->sequence = sequence;
  p->is_modified = is_modified;

  for (i = 0; i < PH_SETTLING_TIME_HISTORY_N; i++) {
    p->st_time[i] = (st_time != NULL) ? st_time[i] : 0.0;
  }

  return p;
}


void ph_routing_entry_dispose(ph_routing_entry_t p) { 
  if (p == NULL) {
    ph_err_msg("ph_routing_entry_dispose() : NULL input");
    return;
  }
  
  ph_vp_range_dispose(p->dest);
  ph_free(p);
}



static ph_routing_entry_t ph_routing_entry_dup(ph_routing_entry_t p) {
  return ph_routing_entry_new(p->dest->lower, p->dest->upper, p->conn, p->metric, p->sequence, p->is_modified, p->st_time, p->addr_type);
}



PH_DEFINE_LIST_FUNC_BODY(ph_routing_entry)


ph_routing_entry_list_t ph_routing_entry_list_new2(ph_vp_set_t ph_vp_set) {
  ph_routing_entry_list_t list;
  ph_vp_range_cell_t p;

  list = ph_routing_entry_list_new(0);

  assert(ph_vp_set != NULL);
  for (p = ph_vp_set->head; p != NULL; p = p->next) {
    ph_routing_entry_list_enqueue(list, ph_routing_entry_new(p->self->lower, p->self->upper, NULL, 0, 0, 0, NULL, unknown));
  }
  return list;
}


ph_routing_event_t ph_routing_event_new(ph_routing_event_type_t type, ph_conn_t from_conn, ph_routing_entry_list_t entry_list) {
  ph_routing_event_t p;

  p = (ph_routing_event_t)ph_malloc(sizeof(struct ph_routing_event));
  p->type = type;
  p->from_conn = from_conn;
  p->entry_list = entry_list;

  if (p->from_conn != NULL) {
    ph_conn_lock(p->from_conn);
  }
  return p;
}


void ph_routing_event_dispose(ph_routing_event_t p) { 
  if (p == NULL) {
    ph_err_msg("ph_routing_event_dispose() : NULL input");
    return;
  }

  if (p->from_conn != NULL) {
    ph_conn_unlock(p->from_conn);
  }

  if (p->entry_list != NULL) {
    ph_routing_entry_list_dispose(p->entry_list);
  }
  ph_free(p);
}


PH_DEFINE_LIST_FUNC_BODY(ph_routing_event)


#ifdef PH_ROUTING_ENABLE_DOMINATING

ph_routing_neigh_t ph_routing_neigh_new(ph_conn_t conn) {
  ph_routing_neigh_t p;
  int i;

  p = (ph_routing_neigh_t)ph_malloc(sizeof(struct ph_routing_neigh));
  p->addr = NULL;
  p->conn = conn;
  p->local = ph_vp_set_make_empty();
  for (i = 0; i < 2; i++) {
    p->neigh[i] = ph_vp_set_make_empty();
    p->p_neigh[i] = ph_vp_set_make_empty();
  }
  gettimeofday(&p->time, NULL);
  return p;
}



void ph_routing_neigh_dispose(ph_routing_neigh_t p) { 
  int i;

  if (p == NULL) {
    ph_err_msg("ph_routing_neigh_dispose() : NULL input");
    return;
  }

  if (p->addr != NULL) {
    ph_addr_dispose(p->addr);
  }

  ph_vp_set_dispose(p->local);
  for (i = 0; i < 2; i++) {
    ph_vp_set_dispose(p->neigh[i]);
    ph_vp_set_dispose(p->p_neigh[i]);
  }
  ph_free(p);
}



PH_DEFINE_LIST_FUNC_BODY(ph_routing_neigh)



static int is_same_addr(ph_routing_neigh_t neigh, void * arg) {
  ph_addr_t addr = (ph_addr_t)arg;

  return ((neigh->addr != NULL) && 
	  (ph_streq(neigh->addr->host, "xxxx") == 0) && // not accepted connection
	  (ph_streq(neigh->addr->host, addr->host)) &&
	  (ph_streq(neigh->addr->service, addr->service)));
}

ph_routing_neigh_t ph_routing_neigh_list_find_same_addr(ph_routing_neigh_list_t neigh_list, ph_addr_t addr) {
  return ph_routing_neigh_list_find(neigh_list, &is_same_addr, (void *)addr);
}

int ph_routing_neigh_is_same_conn(ph_routing_neigh_t p, void * arg1) {
  assert(arg1 != NULL);
  return (p->conn == (ph_conn_t)arg1);
}


#endif

ph_routing_t ph_routing_new() {
  ph_routing_t routing;

  routing = (ph_routing_t)ph_malloc(sizeof(struct ph_routing));
  routing->entry_list = ph_routing_entry_list_new(0);
  ph_routing_entry_list_insert(routing->entry_list, ph_routing_entry_new(PH_MIN_VP, PH_MAX_VP, NULL, PH_MAX_METRIC, 1L, 0, NULL, unknown));
  
  routing->event_list = ph_routing_event_list_new(1);

#ifdef PH_ROUTING_ENABLE_DOMINATING
  routing->is_gateway = 1;
  routing->neigh_list = ph_routing_neigh_list_new(0);
#endif

  ph_pthread_mutex_init(&routing->mp, NULL);

  return routing;
}


void ph_routing_dispose(ph_routing_t routing) {
  if (routing == NULL) {
    ph_err_msg("ph_routing_dispose() : NULL input");
    return;
  }
  ph_routing_entry_list_dispose(routing->entry_list);
  ph_routing_event_list_dispose(routing->event_list);
#ifdef PH_ROUTING_ENABLE_DOMINATING
  ph_routing_neigh_list_dispose(routing->neigh_list);
#endif
  ph_pthread_mutex_destroy(&routing->mp);
  ph_free(routing);
}


void ph_routing_print(ph_routing_t routing) {
  ph_routing_entry_cell_t p;

  assert(routing != NULL);
  
  ph_pthread_mutex_lock(&routing->mp);
  ph_debug_print2("\n[routing\n");

  for (p = routing->entry_list->head; p != NULL; p = p->next) {
    ph_debug_print("[" PH_LLD "," PH_LLD "], conn=(%p), metric=%d, sequence=%ld\n", p->self->dest->lower, p->self->dest->upper, p->self->conn, p->self->metric, p->self->sequence);
  }
  
  ph_debug_print2("]\n");
  ph_pthread_mutex_unlock(&routing->mp);
}


static int is_mergeable(ph_routing_entry_t p, ph_routing_entry_t q) {
  return ((p->metric == q->metric) && 
	  (p->conn == q->conn) &&
	  ((p->metric == 0) || (p->sequence == q->sequence)));
}


static void ph_routing_entry_list_merged_enqueue(ph_routing_entry_list_t list, ph_routing_entry_t p) {
  ph_routing_entry_t q;

  if ((ph_routing_entry_list_is_empty(list) == 0) && (is_mergeable(p, list->tail->self))) {
    int i;

    q = list->tail->self;

    q->dest->upper = p->dest->upper; 
    q->is_modified = (p->is_modified) || (q->is_modified);
    
    for (i = 0; i < PH_SETTLING_TIME_HISTORY_N; i++) {
      q->st_time[i] = ph_max(p->st_time[i], q->st_time[i]);
    }
    
    if (p->metric == 0) {
      q->sequence = ph_max(p->sequence, q->sequence);
    }
    ph_routing_entry_dispose(p);
  } else {
    ph_routing_entry_list_enqueue(list, p);
  }
}


static void ph_routing_update_entry_list_sub_assume_vps(ph_routing_entry_t p, ph_routing_entry_t q, int * is_inc_update_p, double * inc_update_time_p) {
  if (q == NULL) {
    return;
  }

  p->conn = NULL;
  p->metric = 0;
  p->sequence += (p->sequence % 2L == 0) ? 2L : 1L;
  p->is_modified = 1;
 (*is_inc_update_p) = 1;
 (*inc_update_time_p) = 0.0;
}

static void ph_routing_update_entry_list_sub_release_vps(ph_routing_entry_t p, ph_routing_entry_t q, int * is_inc_update_p, double * inc_update_time_p) {
  if (q == NULL) {
    return;
  }

  p->conn = NULL;
  p->metric = PH_MAX_METRIC;
  p->sequence += (p->sequence % 2L == 0) ? 1L : 0;
  p->is_modified = 1;
 (*is_inc_update_p) = 1;
 (*inc_update_time_p) = 0.0;
}

static void ph_routing_update_entry_list_sub_fail(ph_routing_entry_t p, ph_routing_entry_t q, ph_routing_event_t event, int * is_inc_update_p, double * inc_update_time_p, int is_gateway) {
  if (p->conn != event->from_conn) {
    return;
  }
  p->conn = NULL;
  p->metric = PH_MAX_METRIC;
  p->sequence += (p->sequence % 2L == 0) ? 1L : 0;
#ifdef PH_ROUTING_ENABLE_DOMINATING
  if (is_gateway) {
#endif
    p->is_modified = 1;
    (*is_inc_update_p) = 1;
    (*inc_update_time_p) = 0.0;
#ifdef PH_ROUTING_ENABLE_DOMINATING
  }
#endif
}


static void  ph_routing_update_entry_list_sub_update(ph_routing_entry_t p, ph_routing_entry_t q, ph_routing_event_t event, int * is_inc_update_p, double * inc_update_time_p, int is_gateway) {
  if (q == NULL) {
    return;
  }

  if ((q->sequence > p->sequence) || ((p->sequence == q->sequence) && (q->metric + 1 < p->metric))) {
    int i;

    /* TODO: entry $B$4$H$K(B settling time $B$,7hDj$9$Y$-(B */

#ifdef PH_ROUTING_ENABLE_DOMINATING
    if ((is_gateway) || (q->metric == 0)) { 
#endif
      (*is_inc_update_p) = 1;
#ifdef PH_ROUTING_ENABLE_DOMINATING
  }
#endif

    if (q->sequence > p->sequence) {
      double t = 0.0;

      gettimeofday(&p->first, NULL);
      for (i = 0; i < PH_SETTLING_TIME_HISTORY_N; i++) {
	t += p->st_time[i];
      }

#ifdef PH_ROUTING_ENABLE_DOMINATING
    if ((is_gateway) || (q->metric == 0)) { 
#endif
	(*inc_update_time_p) = ph_max((*inc_update_time_p), t / (double)PH_SETTLING_TIME_HISTORY_N);
#ifdef PH_ROUTING_ENABLE_DOMINATING
      }
#endif
      
      p->st_time[PH_SETTLING_TIME_HISTORY_N-1] = 0.0;
      for (i = 0; i < PH_SETTLING_TIME_HISTORY_N-1; i++) {
	p->st_time[i] = p->st_time[i+1];
      }
      
      //fprintf(stderr, "update (sequence)\t st=%f\n", (*inc_update_time_p));
    } else {
      struct timeval best;

      gettimeofday(&best, NULL);
      p->st_time[PH_SETTLING_TIME_HISTORY_N-1] = (double)(best.tv_sec - p->first.tv_sec) + ((double)(best.tv_usec - p->first.tv_usec))/1000000.0;

#ifdef PH_ROUTING_ENABLE_DOMINATING
    if ((is_gateway) || (q->metric == 0)) { 
#endif
	(*inc_update_time_p) = 0.0;
#ifdef PH_ROUTING_ENABLE_DOMINATING
      }
#endif
      //fprintf(stderr, "update (metric)\t st=%f\n", p->st_time[PH_SETTLING_TIME_HISTORY_N-1]);
    }
     
    if (p->metric == 0) {    
      p->sequence = q->sequence + ((q->sequence % 2L == 0) ? 2L : 1L);
    }  else {
      p->metric = (q->metric + 1 < PH_MAX_METRIC) ? q->metric + 1 : PH_MAX_METRIC;
      p->conn = (q->metric + 1 < PH_MAX_METRIC) ? event->from_conn : NULL;
      p->sequence = q->sequence;
    }
    p->is_modified = 1;
  }
}



static void ph_routing_update_entry_list_sub(ph_routing_entry_list_t list, ph_routing_entry_t _p, ph_vp_t lower, ph_vp_t upper, ph_routing_entry_t q, ph_routing_event_t event, int * is_inc_update_p, double * inc_update_time_p, int is_gateway) {
  ph_routing_entry_t p;
  
  p = ph_routing_entry_dup(_p);
  p->dest->lower = lower;
  p->dest->upper = upper;

  assert(lower <= upper);
    
  switch (event->type) {
  case assume_vps:
    ph_routing_update_entry_list_sub_assume_vps(p, q, is_inc_update_p, inc_update_time_p);
    break;
  case release_vps:
    ph_routing_update_entry_list_sub_release_vps(p, q, is_inc_update_p, inc_update_time_p);
    break;
  case fail:
    ph_routing_update_entry_list_sub_fail(p, q, event, is_inc_update_p, inc_update_time_p, is_gateway);
    break;
  case update:
    ph_routing_update_entry_list_sub_update(p, q, event, is_inc_update_p, inc_update_time_p, is_gateway);
    break;
 default:
    ph_err_quit("ph_routing_update_entry_list_sub() : no matching");
    break;
  }
  
  ph_routing_entry_list_merged_enqueue(list, p);
}


static ph_routing_entry_list_t ph_routing_update_entry_list(ph_routing_entry_list_t p, ph_routing_event_t event, int * is_inc_update_p, double * inc_update_time_p, int is_gateway) {
  ph_routing_entry_list_t r;
  ph_routing_entry_cell_t i, j;
  ph_vp_t l, u;


  /*
  fprintf(stderr, "\n ----------start---------\n");
  fprintf(stderr, "i=");
  for (i = p->head; i != NULL; i = i->next) {
    fprintf(stderr, "[%lld, %lld]=%d ",  i->self->dest->lower,  i->self->dest->upper, i->self->metric);
  }
  fprintf(stderr, "\n");

  fprintf(stderr, "j=");
  for (j =  event->entry_list->head; j != NULL; j = j->next) {
    fprintf(stderr, "[%lld, %lld]=%d ",  j->self->dest->lower,  j->self->dest->upper, j->self->metric);
  }
  fprintf(stderr, "\n");
  */

  r = ph_routing_entry_list_new(0);
  
  i = p->head;
  j = event->entry_list->head;
  l = (i != NULL) ? i->self->dest->lower : 0LL;
  u = (i != NULL) ? i->self->dest->upper : 0LL;
  while (i != NULL) {
    if (j == NULL) {
      ph_routing_update_entry_list_sub(r, i->self, l, u, NULL, 
				       event, is_inc_update_p, inc_update_time_p, is_gateway);
      i = i->next;
      l = (i != NULL) ? i->self->dest->lower : 0LL;
      u = (i != NULL) ? i->self->dest->upper : 0LL;
      continue;
    } 

    if (u < j->self->dest->lower) { 	/* [l,u] is lesser than j */
      ph_routing_update_entry_list_sub(r, i->self, l, u, NULL, 
				       event, is_inc_update_p, inc_update_time_p, is_gateway);

      l = u + ph_get_ph_vp_t(1);
      u = i->self->dest->upper;

      if (l > u) {
	i = i->next;
	l = (i != NULL) ? i->self->dest->lower : 0LL;
	u = (i != NULL) ? i->self->dest->upper : 0LL;
      }
      continue;
    }
    
    if ((l > j->self->dest->upper)) {	/* j is lesser than [l,u] */
      j = j->next;
      continue;
    } 

    /* Then, i and j are intersected */

    if ((j->self->dest->lower <= l) && (l <= j->self->dest->upper)) {
      if (j->self->dest->upper >= u) {
	ph_routing_update_entry_list_sub(r, i->self, l, u, j->self, 
					 event, is_inc_update_p, inc_update_time_p, is_gateway);
	i = i->next;
	l = (i != NULL) ? i->self->dest->lower : 0LL;
	u = (i != NULL) ? i->self->dest->upper : 0LL;
      } else {
	ph_routing_update_entry_list_sub(r, i->self, l, j->self->dest->upper, j->self, 
					 event, is_inc_update_p, inc_update_time_p, is_gateway);
	l = j->self->dest->upper + ph_get_ph_vp_t(1);
      }
      continue;
    }
    
    if ((l <= j->self->dest->lower) && (j->self->dest->lower <= u)) {
      u = j->self->dest->lower - ph_get_ph_vp_t(1);
      continue;
    }
  
  } 

  /*
  fprintf(stderr, "r=");
  for (j =  r->head; j != NULL; j = j->next) {
    fprintf(stderr, "[%lld, %lld]=%d ",  j->self->dest->lower,  j->self->dest->upper, j->self->metric);
  }
  fprintf(stderr, "\n----------end---------\n");
  */
  return r;
}


#ifdef PH_ROUTING_ENABLE_DOMINATING


static void calculate_neigh(ph_routing_neigh_list_t neigh_list, ph_routing_event_t event) {
  ph_routing_neigh_t neigh;
  ph_routing_entry_cell_t p;

  if (event->from_conn == NULL) {
    return;
  }

  if ((neigh = ph_routing_neigh_list_find(neigh_list, &ph_routing_neigh_is_same_conn, event->from_conn)) == NULL) {
    /* if a new link is added */
    neigh = ph_routing_neigh_new(event->from_conn);
    ph_routing_neigh_list_insert(neigh_list, neigh);
  }

  /* remove a neighbor whose link gets broken */
  if (event->type == fail) {
    ph_routing_neigh_list_delete(neigh_list, neigh);

    /* TODO: $BK\Ev$K$3$N;~(B neigh$B$r%j%9%H$+$i:o=|$9$Y$-!)(B */
    return;
  }

  for (p = event->entry_list->head; p != NULL; p = p->next) {
    switch (p->self->metric) {
    case 0:	 /* neigh's local vp */
      ph_vp_set_add_range(&neigh->local, p->self->dest->lower, p->self->dest->upper);
      break;
    case 1:      /* neigh's direct neighbor */
      switch (p->self->addr_type) {
      case tcp:
      case unknown:
	ph_vp_set_add_range(&neigh->neigh[tcp], p->self->dest->lower, p->self->dest->upper);
	ph_vp_set_remove_range(&neigh->neigh[ssh], p->self->dest->lower, p->self->dest->upper);
	break;
      case ssh:
	ph_vp_set_remove_range(&neigh->neigh[tcp], p->self->dest->lower, p->self->dest->upper);
	ph_vp_set_add_range(&neigh->neigh[ssh], p->self->dest->lower, p->self->dest->upper);
	break;
      default:
	ph_err_quit("calculate_neigh() : no matching");
	break;
      }
      break;
    default:     /* remove obsolete information */
      ph_vp_set_remove_range(&neigh->neigh[tcp], p->self->dest->lower, p->self->dest->upper);
      ph_vp_set_remove_range(&neigh->neigh[ssh], p->self->dest->lower, p->self->dest->upper);
      break;
    }
  }
}

static int is_direct_neigh_sub(ph_routing_neigh_t p, ph_routing_neigh_t q, ph_addr_type_t type) {
  return ((ph_vp_set_is_sub(p->local, q->neigh[type])) && 
	  (ph_vp_set_is_sub(q->local, p->neigh[type])));
}


static int is_direct_neigh(ph_routing_neigh_t p, ph_routing_neigh_t q) {
  return ((is_direct_neigh_sub(p, q, tcp)) || (is_direct_neigh_sub(p, q, ssh)));
}


static void calculate_gateway(ph_routing_t routing) {
  int n = 0;
  ph_routing_neigh_cell_t p, q;
  
  /*
  {
    for (p = routing->neigh_list->head; p != NULL; p = p->next) {
      ph_vp_range_cell_t v;
      
      fprintf(stderr, "[%lld] ", p->self->local->head->self->lower);
      fprintf(stderr, "tcp= ");
      for (v = p->self->neigh[0]->head; v != NULL; v = v->next) {
	fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s",
		v->self->lower, v->self->upper, ((v->next != NULL) ? "," : ""));
      }
      fprintf(stderr, "\n");
      
      fprintf(stderr, "ssh= ");
      for (v = p->self->neigh[1]->head; v != NULL; v = v->next) {
	fprintf(stderr, "("  PH_LLD "," PH_LLD ")%s",
		v->self->lower, v->self->upper, ((v->next != NULL) ? "," : ""));
      }
      fprintf(stderr, "\n");
    }
  }
  */
  
  for (p = routing->neigh_list->head; p != NULL; p = p->next) {
    for (q = p->next; q != NULL; q = q->next) {
      if (ph_vp_set_is_equal(p->self->local, q->self->local)) {
	continue;
      }

      if (is_direct_neigh(p->self, q->self) == 0) {
	/*
	fprintf(stderr, "(gateway) p=%lld, q=%lld\n", p->self->local->head->self->lower,
		q->self->local->head->self->lower);
	*/
	n = 1;
      }
    }
  }
  routing->is_gateway = n;
}
#endif


static void ph_routing_handle_repair(ph_routing_t routing, ph_routing_event_t event) {
  event->from_conn->is_repaired = 1;
}


static void ph_routing_handle_finalize(ph_routing_t routing, ph_routing_event_t event, int * is_inc_update_p, double * inc_update_time_p) {
  (*is_inc_update_p) = 1;
  (*inc_update_time_p) = 0.0;
}


void ph_routing_handle_event(ph_routing_t routing, ph_routing_event_t event, int * is_inc_update_p, double * inc_update_time_p) {
  switch (event->type) {
  case assume_vps:
  case release_vps:
  case fail:
  case update: {
    ph_routing_entry_list_t r;
    
    //fprintf(stderr, "update\n");
    ph_pthread_mutex_lock(&routing->mp);

    /* swap entry_list with new entry_list */
    r = ph_routing_update_entry_list(routing->entry_list, event, is_inc_update_p, inc_update_time_p, 
#ifdef PH_ROUTING_ENABLE_DOMINATING
				     routing->is_gateway
#else
				     0
#endif
				     );
    ph_routing_entry_list_dispose(routing->entry_list);
    routing->entry_list = r;

#ifdef PH_ROUTING_ENABLE_DOMINATING
    /*
    fprintf(stderr, "accessing neigh_list (start)\n");
    calculate_neigh(routing->neigh_list, event);
    calculate_gateway(routing);
    fprintf(stderr, "accessing neigh_list (end)\n");
    */
#endif

    ph_pthread_mutex_unlock(&routing->mp);
    break;
  }
  case repair:
    ph_routing_handle_repair(routing, event);
    break;
  case finalize:
    ph_routing_handle_finalize(routing, event, is_inc_update_p, inc_update_time_p);
    break;
  default:
    ph_err_quit("ph_routing_handle_event() : no matching");
    break;
  }
}


double ph_routing_wait_event_list(ph_routing_t routing, double wait_time) {
#ifndef __PH_WIN32__  
  struct timespec abstime;
  struct timeval start, end;
#else
  DWORD start, end;
#endif

#ifndef __PH_WIN32__  
  {
    gettimeofday(&start, NULL);

    abstime.tv_sec = start.tv_sec + ((wait_time > 1.0) ? (int)wait_time : 1);
    abstime.tv_nsec = start.tv_usec / 1000;
  }
#else
  start = timeGetTime();
#endif
  
  ph_pthread_mutex_lock(&routing->event_list->mp);
  if (ph_routing_event_list_is_empty_opt(routing->event_list, 0)) {
#ifndef __PH_WIN32__
    pthread_cond_timedwait(&routing->event_list->i_cond, &routing->event_list->mp, &abstime);
#else 
    ph_cond_timedwait(&routing->event_list->i_cond, &routing->event_list->mp, (DWORD)(wait_time * 1000.0));
#endif
  }
  ph_pthread_mutex_unlock(&routing->event_list->mp);

#ifndef __PH_WIN32__
  gettimeofday(&end, NULL);
  return (double)(end.tv_sec - start.tv_sec) + ((double)(end.tv_usec - start.tv_usec))/1000000.0;
#else
  end = timeGetTime();
  return (double)(end - start) * 1000.0;
#endif
}


static ph_conn_t get_conn(ph_routing_entry_list_t entry_list, ph_vp_t vp) {
  ph_routing_entry_cell_t p;
  
  for (p = entry_list->head; p != NULL; p = p->next) {
    if (ph_vp_range_is_member(p->self->dest, vp)) {
      return p->self->conn;
    }
  }

  assert(0);
  return NULL;
}


static void ph_routing_route_msg_sub(ph_routing_t routing, ph_msg_t msg) {
  assert(msg != NULL);
  
  if (msg->type == meta) {
    return;
  }

  ph_pthread_mutex_lock(&msg->mp);
  msg->to_conn = get_conn(routing->entry_list, msg->dest);
  ph_pthread_mutex_unlock(&msg->mp);
}


void ph_routing_route_enqueue_msg(ph_routing_t routing, ph_msg_list_t msg_list, ph_msg_t msg) {
  ph_pthread_mutex_lock(&routing->mp);
  ph_routing_route_msg_sub(routing, msg);
  ph_msg_list_enqueue(msg_list, msg);
  ph_pthread_mutex_unlock(&routing->mp);
}


void ph_routing_route_msg_list(ph_routing_t routing, ph_msg_list_t msg_list) {
  ph_msg_cell_t p;

  ph_pthread_mutex_lock(&routing->mp);
  ph_pthread_mutex_lock(&msg_list->mp);
  for (p = msg_list->head; p != NULL; p = p->next) {
    ph_routing_route_msg_sub(routing, p->self);
  }
  ph_pthread_cond_broadcast(&msg_list->i_cond);
  ph_pthread_mutex_unlock(&msg_list->mp);
  ph_pthread_mutex_unlock(&routing->mp);
}


void ph_routing_increment_sequences(ph_routing_t routing) {
  ph_routing_entry_cell_t p;

  ph_pthread_mutex_lock(&routing->mp);
  for (p = routing->entry_list->head; p != NULL; p = p->next) {
    if (p->self->metric == 0) {
      assert(p->self->sequence % 2L == 0);
      p->self->sequence += 2L;
    }
  }
  ph_pthread_mutex_unlock(&routing->mp);
}


static int get_nelm(ph_routing_t routing, int is_full, int is_finalizing) {
  int nelm = 0;
  ph_routing_entry_cell_t p;
  
  for (p = routing->entry_list->head; p != NULL; p = p->next) {
    if ((is_full) || (p->self->is_modified) || (is_finalizing)) {
      nelm++;
    }
  }
  return nelm;
}

/*
  TODO: non-gateway $B$O$I$3$^$G@53N$J>pJs$rEAHB$5$;$k$+!)(B 
  (0hop, 1hop, $B$=$l0J30$G(Bmerge $B$7$F$b$h$$5$$,$9$k$,(B, sequence number $B$b9M$($k$H!)(B)
 */
char * ph_routing_create_msg(ph_routing_t routing, int * len_p, int is_full, int is_finalizing) {
  int PH_ENTRY_SIZE = 100;
  int nelm;
  char * body, * s;
  ph_routing_entry_cell_t p;
  
  ph_pthread_mutex_lock(&routing->mp);

  if ((nelm = get_nelm(routing, is_full, is_finalizing)) == 0) {
    ph_pthread_mutex_unlock(&routing->mp);
    return NULL;
  }

  (*len_p) = (nelm+1) * PH_ENTRY_SIZE;

  body = (char *)ph_calloc((*len_p), sizeof(char));
  ph_mzero(body, (*len_p));
  
  ph_snprintf(body, PH_ENTRY_SIZE, "%d", nelm);
  
  s = body + PH_ENTRY_SIZE;
  for (p = routing->entry_list->head; p != NULL; p = p->next) {
    if ((is_full) || (p->self->is_modified) || (is_finalizing)) {
      ph_snprintf(s, PH_ENTRY_SIZE, PH_LLD " " PH_LLD " %d %d %d", 
		  p->self->dest->lower, 
		  p->self->dest->upper, 
		  ((is_finalizing) ? PH_MAX_METRIC : p->self->metric), 
		  p->self->sequence,
		  ((p->self->conn != NULL) ? (int)ph_conn_get_addr_type(p->self->conn) : (int)unknown));
      
      if (is_full == 0) {
	p->self->is_modified = 0;
	// is_modified becomes 0 only if modfication is broadcasted to all the neighbors
      }
      s += PH_ENTRY_SIZE;
    }
  }
  
  ph_pthread_mutex_unlock(&routing->mp);

  return body;
}


ph_routing_entry_list_t ph_routing_analyze_msg(char * entries) {
  int PH_ENTRY_SIZE = 100;
  int nelm, i;
  ph_routing_entry_list_t entry_list;
  char * s;

  entry_list = ph_routing_entry_list_new(0);

  if (entries == NULL) {
    return entry_list;
  }
 
  sscanf(entries, "%d", &nelm);
  
  for (i = 0, s = entries + PH_ENTRY_SIZE; i < nelm; i++, s+= PH_ENTRY_SIZE) {
    ph_vp_t lower, upper;
    int metric, sequence;
    ph_addr_type_t addr_type;

    sscanf(s, PH_LLD " " PH_LLD " %d %d %d", &lower, &upper, &metric, &sequence, (int *)&addr_type);
    ph_routing_entry_list_enqueue(entry_list, ph_routing_entry_new(lower, upper, NULL, metric, sequence, 0, NULL, addr_type));
  }

  return entry_list;
}
















