/**************************************
 * $Author: kaneda $
 * $Date: 2003/03/13 12:51:22 $
 **************************************/

/* tcp.c */
#include "ph_conn.h"

#ifndef __PH_WIN32__
#include "sys/socket.h"
#endif

ph_tcp_attr_t ph_tcp_attr_new(int fd) {
  ph_tcp_attr_t p;
  int oldflags;
  int one = 1;

  p = (ph_tcp_attr_t)ph_malloc(sizeof(struct ph_tcp_attr));
  p->fd = fd;

#ifndef __PH_WIN32__
  oldflags = ph_fcntl(p->fd, F_GETFD, 0);
  ph_fcntl(p->fd, F_SETFD, oldflags | FD_CLOEXEC);

  ph_setsockopt(p->fd,		/* socket affected */
		IPPROTO_TCP,	/* set option at TCP level */
		TCP_NODELAY,	/* name of option */
		(char *)&one,	/* the cast is historical cruft */
		sizeof(int));	/* length of option value */
#endif


  return p;
}


void ph_tcp_attr_dispose(void * p) {
  ph_tcp_attr_t attr = (ph_tcp_attr_t)p;

  if (attr == NULL) {
    //ph_err_msg("ph_tcp_attr_dispose() : NULL input");
    return;
  }

  ph_free(attr);
}

#ifndef __PH_WIN32__

static void connect_alarm(int signo) {
  return;		/* just interrupt the connect() */
}

static int connect_timeo(int sockfd, const struct sockaddr *saptr, socklen_t salen, int nsec) {
  ph_sigfunc	*sigfunc;
  int		n;

  {
    sigset_t set;
    
    sigemptyset(&set);
    sigaddset(&set, SIGALRM);
    ph_pthread_sigmask(SIG_UNBLOCK, &set, NULL);
  }

  sigfunc = ph_signal(SIGALRM, connect_alarm);
  if (alarm(nsec) != 0) {
    //ph_err_msg("connect_timeo: alarm was already set");
  }
  
  if ((n = connect(sockfd, (struct sockaddr *) saptr, salen)) < 0) {
    if (errno == EINTR) {
      errno = ETIMEDOUT;
    }
  }

  alarm(0);			/* turn off the alarm */
  ph_signal(SIGALRM, sigfunc);	/* restore previous signal handler */
  
  {
    sigset_t set;
    
    sigemptyset(&set);
    sigaddset(&set, SIGALRM);
    ph_pthread_sigmask(SIG_BLOCK, &set, NULL);
  }

  return n;
}

static int connect_nonb(int sockfd, const struct sockaddr *saptr, socklen_t salen, int nsec) {
  int			flags, n, error;
  socklen_t		len;
  fd_set		rset, wset;
  struct timeval	tval;

  flags = ph_fcntl(sockfd, F_GETFL, 0);
  ph_fcntl(sockfd, F_SETFL, flags | O_NONBLOCK);

  error = 0;

  if ((n = connect(sockfd, (struct sockaddr *) saptr, salen)) < 0) {
    if ((errno != 0) && (errno != EINPROGRESS)) {
      return -1;
    }
  }

  /* Do whatever we want while the connect is taking place. */
  
  if (n == 0) {
    goto done;	/* connect completed immediately */
  }
  
  FD_ZERO(&rset);
  FD_ZERO(&wset);
  FD_SET(sockfd, &rset);
  FD_SET(sockfd, &wset);
  tval.tv_sec = nsec;
  tval.tv_usec = 0;
  
  if ((n = select(sockfd+1, &rset, &wset, NULL, nsec ? &tval : NULL)) == 0) {
    /* timeout */
    errno = ETIMEDOUT;
    return -1;
  }

  if (FD_ISSET(sockfd, &rset) || FD_ISSET(sockfd, &wset)) {
    len = sizeof(error);
    if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, &error, &len) < 0) {
      return -1;			/* Solaris pending error */
    }
  } else {
    ph_err_quit("select error: sockfd not set");
  }
  
 done:
  ph_fcntl(sockfd, F_SETFL, flags);	/* restore file status flags */
  
  if (error) {
    errno = error;		/* just in case */
    return -1;
  }
  return 0;
}

#endif

static int ph_tcp_connect(const char * host, const char * serv) {
  int			sockfd;
  struct hostent * 	hp;
  struct sockaddr_in 	sin;

  sockfd = ph_socket(AF_INET, SOCK_STREAM, 0);
  
  ph_mzero(&sin, sizeof(sin));  
  sin.sin_family = AF_INET;
  sin.sin_port = htons(atoi(serv));
  
  if ((sin.sin_addr.s_addr = inet_addr(host)) == /*(in_addr_t)-1*/ INADDR_NONE) {
    if ((hp = gethostbyname(host)) == NULL) {
#ifndef __PH_WIN32__
      ph_close(sockfd);
#else
      closesocket(sockfd);
#endif
      return -1;
    }
    memcpy(&sin.sin_addr, hp->h_addr, hp->h_length);
  }

#ifndef __PH_WIN32__
  //if (connect(sockfd, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
  //if (connect_timeo(sockfd, (struct sockaddr *)&sin, sizeof(sin), PH_CONNECT_TIMEOUT) == -1) {
  if (connect_nonb(sockfd, (struct sockaddr *)&sin, sizeof(sin), PH_CONNECT_TIMEOUT) == -1) {
      ph_close(sockfd);
    return -1;
  }
#else
  if (connect(sockfd, (struct sockaddr *)&sin, sizeof(sin)) == -1) {
    closesocket(sockfd);
    return -1;
  }
#endif

  return sockfd;
}



void * ph_tcp_attr_connect(ph_addr_t addr) {
  int fd;

  assert(addr != NULL);
  if ((fd = ph_tcp_connect(addr->host, addr->service)) == -1) {
    return NULL;
  }

  if (ph_conn_is_test) { 
    fprintf(stderr, "successfully connected to %s:%s\n", 
	    addr->host, addr->service);
    fflush(stderr);
  }
  return (void *)ph_tcp_attr_new(fd);
}


static int ph_tcp_listen(const char *host, const char *serv, socklen_t *addrlenp) {
  int			listenfd;
  const int		on = 1;
  struct sockaddr_in 	servaddr;


  listenfd = ph_socket(AF_INET, SOCK_STREAM, 0);
  
  ph_mzero(&servaddr, sizeof(servaddr));
  servaddr.sin_family = AF_INET;
  servaddr.sin_addr.s_addr = htonl(INADDR_ANY);
  servaddr.sin_port = htons(atoi(serv));
  
  ph_setsockopt(listenfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));
   
  if (bind(listenfd, (struct sockaddr *)&servaddr, sizeof(servaddr)) < 0) {
    return -1;
  }
  ph_listen(listenfd, PH_LISTENQ);

  if (addrlenp != NULL) {
    *addrlenp = sizeof(servaddr);	/* return size of protocol address */
  }
 
  return(listenfd);
}


void * ph_tcp_attr_listen(ph_addr_t addr) {
  int fd;

  assert(addr != NULL);
  if ((fd = ph_tcp_listen(addr->host, addr->service, NULL)) < 0) {
    return NULL;
  }
  return (void *)ph_tcp_attr_new(fd);
}


void * ph_tcp_attr_accept(void * p, struct sockaddr * cliaddr_p, socklen_t * len_p) {
  ph_tcp_attr_t listen_attr = (ph_tcp_attr_t)p;
  int fd;

  assert(listen_attr != NULL);
  if ((fd = accept(listen_attr->fd, cliaddr_p, len_p)) < 0) {
    return NULL;
  }
  return (void *)ph_tcp_attr_new(fd);
}


static ssize_t sendn(int fd, void * buf, size_t len) {
  ssize_t n = 0, m;
 
  while (((m = send(fd, (char *)buf + n, len - n, 0)) > 0) && ((n += m) < len)) {
  }
  // (errno == EINTR) $B$J$i$P!)(B

  return n;
}

ssize_t ph_tcp_attr_send(void * p, void * buf, size_t len) {
  ph_tcp_attr_t attr = (ph_tcp_attr_t)p;
  
  assert(attr != NULL);
  return sendn(attr->fd, buf, len);
}


static ssize_t recvn(int fd, void * buf, size_t len) {
  ssize_t n = 0, m;
 
  while (((m = recv(fd, (char *)buf + n, len - n, 0)) > 0) && ((n += m) < len)) {
  }

  return n;
}

ssize_t ph_tcp_attr_recv(void * p, void * buf, size_t len) {
  ph_tcp_attr_t attr = (ph_tcp_attr_t)p;
  
  assert(attr != NULL);
  return recvn(attr->fd, buf, len);
}

void ph_tcp_attr_close(void * p) {
  ph_tcp_attr_t attr = (ph_tcp_attr_t)p;

  assert(attr != NULL); 

  assert(attr->fd >= 0);
#ifndef __PH_WIN32__
  shutdown(attr->fd, 2);
  ph_close(attr->fd);
#else
  closesocket(attr->fd);
#endif
}


int ph_tcp_attr_fd_isset(void * p, fd_set * fdset_p, int type) {
  ph_tcp_attr_t attr = (ph_tcp_attr_t)p;

  assert(attr != NULL); 
  assert(fdset_p != NULL);

  return ((attr->fd >= 0) && (FD_ISSET(attr->fd, fdset_p)));
}


void ph_tcp_attr_fd_set(void * p, fd_set * fdset_p, int type) {
  ph_tcp_attr_t attr = (ph_tcp_attr_t)p;

  assert(attr != NULL);
  assert(fdset_p != NULL);

  FD_SET(attr->fd, fdset_p);
}



#if 0

static int tcp_connect(const char *host, const char *serv) {
  int                   sockfd, n;
  struct addrinfo       hints, *res, *ressave;

  bzero(&hints, sizeof(struct addrinfo));
  hints.ai_family = AF_UNSPEC;
  hints.ai_socktype = SOCK_STREAM;
   
  /* getaddrinfo() is MT unsafe */
  Pthread_mutex_lock(&mp);
  if ((n = getaddrinfo(host, serv, &hints, &res)) != 0) {
    err_quit("tcp_connect error for %s, %s: %s", host, serv, gai_strerror(n));
  }
  Pthread_mutex_unlock(&mp);
  
  ressave = res;
  
  do {
    sockfd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
    if (sockfd < 0) {
      continue; /* ignore this one */
    }
    if (connect(sockfd, res->ai_addr, res->ai_addrlen) == 0) {
      break;            /* success */
    }

    Close(sockfd);      /* ignore this one */
  } while ((res = res->ai_next) != NULL);

  if (res == NULL) {    /* errno set from final connect() */
    //err_sys("tcp_connect error for %s, %s", host, serv);
    sockfd = -1;
  }
  
  /* freeaddrinfo() is MT Safe ??? */
  Pthread_mutex_lock(&mp);
  freeaddrinfo(ressave);
  Pthread_mutex_unlock(&mp);
  
  return sockfd;
}
#endif
