#include "viewer.h"
#include "math.h"


/* get the color value from its name */
unsigned long get_color(Display* display, char* color_name) {
  Colormap 	cmap;
  XColor 	near_color, true_color;
  
  cmap = DefaultColormap(display, 0);
  XAllocNamedColor(display, cmap, color_name, &near_color, &true_color);
  return near_color.pixel;
}

static void window_new(Display* display, Window * window_p, Window parent) {
  Window window;
  
  /* create a new window */
  (*window_p) = XCreateSimpleWindow(display,
				    parent, 			/* parent window */
				    0, 				/* xo */
				    0, 				/* yo */
				    WINDOW_WIDTH,		/* width */
				    WINDOW_HEIGHT, 		/* height */
				    3,				/* border_width */
				    WhitePixel(display,0), 	/* border_color */
				    BlackPixel(display,0));	/* ground_color */
  window = (*window_p);
  
  /* setup the window attribute */
  {
    XSetWindowAttributes attr;		/* window attribute */

    attr.backing_store = WhenMapped;
    /* make the window remember the drawn picture */
    XChangeWindowAttributes(display, window, CWBackingStore, &attr);
  }

  /* setup the window title */
  XStoreName(display, window, "Phoenix Viewer");

  /* wait until actually X server receives the request and opens the window */
  {
    XEvent event;	/* event */
    
    XSelectInput(display, window, ExposureMask);
    /* show the window */
    XMapWindow(display, window);
    
    do{
      XNextEvent(display, &event);
    } while(event.type != Expose);
  }
}

static void gc_new(Display * display, Window window, GC * gc_p) {
  (*gc_p) = XCreateGC(display, 
		      window,
		      0,	/* mask */
		      NULL);	/* XGC Values */
}

void initialize_X(viewer_t viewer) {
  assert(viewer != NULL);

  /* connect to X server */
  if ((viewer->display = XOpenDisplay(NULL)) == NULL) {
    ph_err_quit("XOpenDisplay() failed");
  }

  window_new(viewer->display, &viewer->windows[0], RootWindow(viewer->display,0));
  gc_new(viewer->display, viewer->windows[0],  &viewer->gcs[0]);

  //viewer->font = XLoadFont(viewer->display, "9x15");
  viewer->font = XLoadFont(viewer->display, "6x10");
  XSetFont(viewer->display, viewer->gcs[0], viewer->font);
}


void finalize_X(viewer_t viewer) {
  int i;

  assert(viewer != NULL);

  /* destroy font */
  XUnloadFont(viewer->display, viewer->font);


  /* destroy gc */
  for (i = 0; i < N_GCS; i++) {
    XFreeGC(viewer->display, viewer->gcs[i]);
  }

  /* destroy the window */
  for (i = 0; i < N_WINDOWS; i++) {
    XDestroyWindow(viewer->display, viewer->windows[i]);
  }

  /* disconnect to X server */
  XCloseDisplay(viewer->display); 
}


static void clean(Display * display, Window window, GC gc) {
  /* erase all the picture */
  XSetForeground(display, gc, get_color(display, "black"));
  XFillRectangle(display, window, gc, 0,0, WINDOW_WIDTH, WINDOW_HEIGHT);
}


static void print_all_vps(viewer_t viewer) {
  Display 	* display = viewer->display;
  Window 	window = viewer->windows[0];
  GC 		gc = viewer->gcs[0];
  ph_vp_set_t vp_set = ph_vp_set_make_empty();
  vinfo_cell_t p;
  char s[8192];

  for (p = viewer->vinfo_list->head; p != NULL; p = p->next) {
    ph_vp_set_t tmp;
    tmp = ph_vp_set_make_union(vp_set, p->self->vp_set);
    ph_vp_set_dispose(vp_set);
    vp_set = tmp;
  }

  ph_mzero(s, 8192);
  snprintf(s, 9, "all_vps=");
  ph_vp_set_vps2str_for_print(vp_set, s);

  XSetForeground(display, gc, get_color(display, "green"));
  XDrawString(display, window, gc, 
	      20, 20,
	      s, ph_strlen(s));
  
  ph_vp_set_dispose(vp_set);
}


static void print_status(viewer_t viewer) {
  Display 	* display = viewer->display;
  Window 	window = viewer->windows[0];
  GC 		gc = viewer->gcs[0];
  vinfo_cell_t p;
  int i;
  
  XSetForeground(display, gc, get_color(display, "green"));
  for (p = viewer->vinfo_list->head, i = 0; p != NULL; p = p->next, i++) {
    char s[8192];
    
    vinfo_vinfo2str(p->self, s, 8192);
    XDrawString(display, window, gc, 20, 50  + i * 20, s, ph_strlen(s));
  }
}

static int is_gateway(vinfo_t vinfo) {
  return (vinfo->comment[0] == 'G');
}

static void draw_graph(viewer_t viewer) {
  Display 	* display = viewer->display;
  Window 	window = viewer->windows[0];
  GC 		gc = viewer->gcs[0];


  vinfo_cell_t p, q;
  int i;
  
  int subx = 256, suby = 300;
  int n_node = vinfo_list_get_nelm(viewer->vinfo_list);
  int R = 100;

  if (n_node == 0) {
    return;
  }

  XSetForeground(display, gc, get_color(display, "green"));
  for (p = viewer->vinfo_list->head, i = 0; p != NULL; p = p->next, i++) {
    p->self->x = subx + (int)(R*cos(2*M_PI*i/n_node));
    p->self->y = suby + (int)(R*sin(2*M_PI*i/n_node));

    if (is_gateway(p->self)) {
      XFillArc(display, window, gc, p->self->x-5 , p->self->y-5, 10, 10, 0, 360*64);
    } else {
      XDrawArc(display, window, gc, p->self->x-5 , p->self->y-5, 10, 10, 0, 360*64);
    }
    
    {
      char s[8192];    
      ph_mzero(s, 8192);
      ph_vp_set_vps2str_for_print(p->self->vp_set, s); 
      XDrawImageString(display, window, gc,
		       subx + (int)(R*cos(2*M_PI*i/n_node) - 10),
		       suby + (int)(R*sin(2*M_PI*i/n_node) + 20),
		       s, ph_strlen(s));
    }
  }


  // $B3F(B vinfo $B$rA4BNA4$NDL?.$,2DG=$G$"$k(B

  
  // $B3F(B $B%[%9%H$N:BI8$r5a$a$k(B

  // $B@\B32DG=$J%N!<%IF1;N$K@~$r0z$/(B

  for (p = viewer->vinfo_list->head; p != NULL; p = p->next) {
    for (q = viewer->vinfo_list->head; q != NULL; q = q->next) {
      int t;

      if (p == q) {
	continue;
      }
      if ((t = vinfo_is_connect(p->self, q->self)) != -1) {
	switch (t) {
	case tcp:
	case unknown:
	  XSetForeground(display, gc, get_color(display, "red"));
	  XDrawLine(display, window, gc, p->self->x, p->self->y, q->self->x, q->self->y); 
	  break;
	case ssh:
	  XSetForeground(display, gc, get_color(display, "blue"));
	XDrawLine(display, window, gc, p->self->x, p->self->y, q->self->x, q->self->y); 
	  break;
	}

      } else if (vinfo_is_routable(p->self, q->self)) {
      	XSetForeground(display, gc, get_color(display, "grey"));
	XDrawLine(display, window, gc, p->self->x, p->self->y, q->self->x, q->self->y); 
      }
    }
  }
}

 void redraw(viewer_t viewer) {
  Display 	* display = viewer->display;
  Window 	window = viewer->windows[0];
  GC 		gc = viewer->gcs[0];

  clean(display, window, gc);

  vinfo_list_sort(viewer->vinfo_list, &vinfo_compare); 

  print_all_vps(viewer);
  print_status(viewer);
  draw_graph(viewer);

  XFlush(display);      
}






