/**************************************
 * $Author: kaneda $
 * $Date: 2002/12/20 10:14:42 $
 **************************************/

/* conn.c */

#include "conn.h"


conn_t conn_new(conn_type_t type, int fd, ph_vp_t vp, task_id_t id) {
  conn_t conn;
  int oldflags;

  conn = (conn_t)Malloc(sizeof(struct conn));
  conn->type = type;
  conn->fd = fd;
  conn->state = OPEN;

  conn->vp = vp;
  conn->id = id;

  oldflags = Fcntl(conn->fd, F_GETFD, 0);
  Fcntl(conn->fd, F_SETFD, oldflags | FD_CLOEXEC);
  
  return conn;
}


void conn_dispose(conn_t conn) { 
  if (conn == NULL) {
    err_msg("conn_dispose() : NULL input");
    return; 
  }

  assert(conn->state == CLOSED);
  
  if (conn->fd >= 0) {
    conn_close(conn);
  }
  Free(conn);
}


conn_t conn_listen(int listen_port) {
  char service[BUFSIZE];

  Mzero(service, BUFSIZE);
  Snprintf(service, BUFSIZE, "%d", listen_port);   
  return conn_new(LISTENING, tcp_listen("0.0.0.0", service, NULL), 0UL, 0);
}


conn_t conn_accept(conn_t listen_conn, ph_vp_t vp) {
  int fd;
  struct sockaddr cliaddr;
  socklen_t len = sizeof(cliaddr);

  assert(listen_conn != NULL);
  if ((fd = accept(listen_conn->fd, &cliaddr, &len)) < 0) {
    return NULL;
  }
  return conn_new(SHELL, fd, vp, 0);
}


static ssize_t sendn(int fd, void * buf, size_t len) {
  ssize_t n = 0, m;
 
  while (((m = send(fd, (char*)buf+n, len - n, 0)) > 0) && ((n += m) < len)) {
  }

  return n;
}


static ssize_t writen(int fd, void * buf, size_t len) {
  ssize_t n = 0, m;
 
  while (((m = write(fd, (char*)buf+n, len - n)) > 0) && ((n += m) < len)) {
  }

  return n;
}

ssize_t conn_send(conn_t conn, void * buf, size_t len) {
  ssize_t n = -1;

  assert(conn != NULL);
  debug_print("conn_send()\t\t\t: (%p)\n", conn);

  switch (conn->type) {
  case SHELL:
    n = sendn(conn->fd, buf, len);
    break;
  case TASK_IN:
    n = writen(conn->fd, buf, len);
    break;
  default:
    err_quit("conn_send() : no matching");
    break;
  }
    
  assert((n <= 0) || (n == len)); 
  return n;
}


static ssize_t recvn(int fd, void * buf, size_t len) {
  ssize_t n = 0, m;
 
  while (((m = recv(fd, (char*)buf+n, len - n, 0)) > 0) && ((n += m) < len)) {
  }

  return n;
}


static ssize_t readn(int fd, char * buf, size_t len) {
  ssize_t n = 0, m;
  
  while (((m = read(fd, (char*)buf+n, len - n)) > 0) && ((n += m) < len)) {
  }

  return n;
}

ssize_t conn_recv(conn_t conn, void * buf, size_t len) {
  ssize_t n = -1;

  assert(conn != NULL);
  debug_print("conn_recv()\t\t\t: (%p)\n", conn);

  switch (conn->type) {
  case SHELL:
    n = recvn(conn->fd, buf, len);
    assert((n <= 0) || (n == len)); 
    break;
  case PHOENIX:
    n = readn(conn->fd, buf, len);
    assert((n <= 0) || (n == len)); 
    break;
  case TASK_OUT:
    n = read(conn->fd, buf, len);
    assert((n <= 0) || (n <= len)); 
    break;
  default:
    err_quit("conn_recv() : no matching");
    break;
  }
    
  return n;
}


void conn_close(conn_t conn) {
  assert(conn != NULL);
  assert(conn->fd >= 0);

  switch (conn->type) {
  case SHELL:
    shutdown(conn->fd, 2);
    Close(conn->fd);
    break;
  case PHOENIX:
  case TASK_IN:
  case TASK_OUT:
    shutdown(conn->fd, 2);
    Close(conn->fd);
    break;
  default:
    err_quit("conn_recv() : no matching");
    break;
  }
  conn->state = CLOSED;
}


DEFINE_LIST_FUNC_BODY(conn) 


static char * print_type(conn_type_t type) {
  switch (type) {
  case LISTENING:
    return "listening";
    break;
  case SHELL:
    return "shell";
    break;
  case PHOENIX:
    return "phoenix";
  case TASK_IN:
    return "task_in";
  case TASK_OUT:
    return "task_out";
    break;
  default:
    err_quit("print_type() : no matching (type=%d)", type);
    break;
  }
  return NULL;
}

static char * print_state(conn_state_t state) {
  switch (state) {
  case OPEN:
    return "open";
    break;
  case CLOSING:
    return "closing";
    break;
  case CLOSED:
    return "closed";
    break;
  default:
    err_quit("print_state() : no matching (state=%d)", state);
    break;
  }

  return NULL;
}

static void conn_print(conn_t conn, void * arg) {
  debug_print("(%p) type=%s, state=%s, fd=%d, vp=%lld, id=%d\n",
	      conn, 
	      print_type(conn->type),
	      print_state(conn->state),
	      conn->fd,
	      conn->vp,
	      (int)conn->id);
}

void conn_list_print(conn_list_t conn_list) {
  assert(conn_list != NULL);

  if (conn_list_is_empty(conn_list)) {
    debug_print2("[conn_list]\n");
  } else {
    debug_print2("[conn_list \n");
    conn_list_map(conn_list, &conn_print, NULL);
    debug_print2("]\n");
  }
}


static int is_same_dest(conn_t conn, ph_vp_t vp, task_id_t id) {
  switch (conn->type) {
  case LISTENING:
  case PHOENIX:
  case TASK_OUT:
    return 0;
    break;
  case SHELL:
    return (conn->vp == vp);
    break;
  case TASK_IN:
    return ((conn->vp == vp) && (conn->id == id));
    break;
  default:
    err_quit("is_same_dest() : no matching");
    break;
  }
  return 0;
}


conn_t conn_list_find_same_dest(conn_list_t conn_list, ph_vp_t vp, task_id_t id) {
  conn_cell_t p;

  assert(conn_list != NULL);
  assert(conn_list->is_sync == 0);

  for (p = conn_list->head; p != NULL; p = p->next) {
    if (is_same_dest(p->self, vp, id)) {
      return p->self;
    }
  }

  return NULL;
}



/* for synchronized iteration */
conn_cell_t conn_list_init_loop(conn_list_t conn_list) {
  assert(conn_list != NULL);
  return conn_list->head;
}


static int is_useless(conn_t conn) {
  assert(conn != NULL);
  return (conn->state == CLOSED);
}


conn_cell_t conn_list_iterate(conn_list_t conn_list, conn_cell_t p) {
  conn_cell_t next;
  
  assert(conn_list != NULL);
  assert(p != NULL);

  next = p->next;
  if (is_useless(p->self)) {
    conn_list_delete(conn_list, p->self);
  }
  return next;
}


